from dataclasses import dataclass

import cv2
import numpy as np

from ...logger import logger


@dataclass
class BoundingBox:
    x1: int
    x2: int
    y1: int
    y2: int


def create_u_net_weight_map(mask, bbox_padding=15, w_c=0.5, w_0=10, sigma=5):
    """

    :param mask:
    :param w_c:
    :param w_0:
    :param sigma:
    :return:
    """

    mask_contours, _ = cv2.findContours(mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)
    logger.debug(f"Number of contours for mask: {len(mask_contours)}")
    bbox = find_bounding_box_for_mask(mask, bbox_padding)
    logger.debug(
        f"Bounding box coordinates for mask: ({bbox[0]}, {bbox[2]}), ({bbox[1]}, {bbox[3]})"
    )
    background_points = find_background_points_in_bbox(mask, *bbox)
    logger.debug(f"Number of background points: {len(background_points)}")
    weights_for_background_points = calculate_weights_for_background_points(
        background_points, mask_contours, w_c, w_0, sigma
    )
    weight_map = create_weight_map(
        background_points, weights_for_background_points, mask, w_c
    )
    return weight_map


def find_bounding_box_for_mask(mask, padding=15):
    mask_bbox = create_mask_bbox(mask)
    contours, _ = cv2.findContours(mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)
    mask_width, mask_height = reversed(list(mask.shape[:2]))
    for contour in contours:
        contour_bbox = create_contour_bbox(contour)
        mask_bbox = update_mask_bbox(
            mask_bbox, contour_bbox, mask_width, mask_height, padding
        )
    return mask_bbox.x1, mask_bbox.x2, mask_bbox.y1, mask_bbox.y2


def create_mask_bbox(mask):
    return BoundingBox(mask.shape[1], 0, mask.shape[0], 0)


def create_contour_bbox(contour):
    x, y, w, h = cv2.boundingRect(contour)
    return BoundingBox(x, x + w, y, y + h)


def update_mask_bbox(mask_bbox, contour_bbox, mask_width, mask_height, padding):
    if mask_bbox.x1 > contour_bbox.x1:
        mask_bbox.x1 = max(contour_bbox.x1 - padding, 0)
    if mask_bbox.x2 < contour_bbox.x2:
        mask_bbox.x2 = min(contour_bbox.x2 + padding, mask_width)
    if mask_bbox.y1 > contour_bbox.y1:
        mask_bbox.y1 = max(contour_bbox.y1 - padding, 0)
    if mask_bbox.y2 < contour_bbox.y2:
        mask_bbox.y2 = min(contour_bbox.y2 + padding, mask_height)
    return mask_bbox


def find_background_points_in_bbox(mask, min_x, max_x, min_y, max_y):
    y, x = np.where(mask[min_y:max_y, min_x:max_x] == 0)
    y += min_y
    x += min_x
    points = list(zip(y, x))
    return points


def calculate_weights_for_background_points(
    background_points, mask_contours, w_c, w_0, sigma
):
    weights = []
    for point in background_points:
        distances_to_contours = sorted(
            calculate_distances_to_contour_from_point(point, mask_contours),
            reverse=True,
        )
        weight = calculate_weight_from_distances(distances_to_contours, w_c, w_0, sigma)
        weights.append(weight)
    return weights


def calculate_distances_to_contour_from_point(point, contours):
    distances_to_contours = []
    for contour in contours:
        open_cv_point = (point[1], point[0])
        distance_from_point_to_contour = cv2.pointPolygonTest(
            contour, open_cv_point, measureDist=True
        )
        distances_to_contours.append(distance_from_point_to_contour)
    return distances_to_contours


def calculate_weight_from_distances(distances, w_c, w_0, sigma):
    if len(distances) == 1:
        p1 = distances[0]
        return calculate_weight(p1, w_c=w_c, w_0=w_0, sigma=sigma)
    else:
        p1, p2, *_ = distances
        return calculate_weight(p1 + p2, w_c=w_c, w_0=w_0, sigma=sigma)


def calculate_weight(distances, w_c=0.5, w_0=10, sigma=5):
    exponent = np.exp(-(np.power(distances, 2) / (2 * sigma ** 2)))
    return w_c + w_0 * exponent


def create_weight_map(points, weights, mask, w_c):
    weight_map = np.zeros_like(mask).astype(np.float32) + w_c
    for point, weight in zip(points, weights):
        weight_map[point] = weight
    return weight_map



from .create_default_istio_gateway_manifest import create_default_istio_gateway_manifest
from .create_seldon_deployment_manifest import create_seldon_deployment_manifest
from .istio import get_istio_ingress_ip, get_istio_ingress_port



import numpy as np

from ....utils.convenience import to_numpy
from ..core import BaseMetric
from .label_statistics import LabelStatisticsForMeanAveragePrecision


class _TensorboardMixin:
    def write_to_tensorboard_on_epoch_end(self, tensorboard_writer, phase, step):
        all_aps = {"mAP": self.mean_average_precision}

        for label in self.labels:
            key = f"{str(phase)}/AP for {label}"
            average_precision = self.average_precision_for_label(label)
            # tensorboard_writer.add_scalar(
            #     key,
            #     average_precision,
            #     step,
            # )
            all_aps[key] = average_precision

        tensorboard_writer.add_scalars(
            f"{str(phase)}/AP for all label NMS: {self.nms_threshold} IOU: {self.iou_threshold}",
            all_aps,
            step,
        )

        self._create_label_statistics_dict()


def mean_average_precision_torchvision_fasterrcnn_transform_fn(
    predictions, targets, auxillary
):
    return {
        "predicted_boxes": [
            to_numpy(prediction["boxes"]) for prediction in predictions
        ],
        "predicted_labels": [
            to_numpy(prediction["labels"]) for prediction in predictions
        ],
        "predicted_scores": [
            to_numpy(prediction["scores"]) for prediction in predictions
        ],
        "target_boxes": [to_numpy(target["boxes"]) for target in targets],
        "target_labels": [to_numpy(target["labels"]) for target in targets],
    }


class MeanAveragePrecision(BaseMetric, _TensorboardMixin):
    def __init__(
        self,
        iou_threshold,
        nms_threshold=None,
        n_classes=None,
        labels=None,
        transform_to_kwargs_fn=mean_average_precision_torchvision_fasterrcnn_transform_fn,
        name=None,
    ):
        assert (n_classes is None) ^ (labels is None)
        super().__init__(name)
        self.nms_threshold = nms_threshold
        self.iou_threshold = iou_threshold
        self.transform_to_kwargs_fn = transform_to_kwargs_fn
        self._n_classes = n_classes
        self._labels = labels
        self._create_label_statistics_dict()

    def transform_to_kwargs(self, predictions, targets, auxillary):
        return self.transform_to_kwargs_fn(predictions, targets, auxillary)

    def _create_label_statistics_dict(self):
        if self._n_classes is not None:
            self.label_statistics_dict = {
                label: LabelStatisticsForMeanAveragePrecision(label)
                for label in range(self._n_classes)
            }

        if self._labels is not None:
            self.label_statistics_dict = {
                label: LabelStatisticsForMeanAveragePrecision(label)
                for label in self._labels
            }

    @property
    def labels(self):
        return list(self.label_statistics_dict.keys())

    @property
    def mean_average_precision(self):
        return np.mean(
            [
                label_stats.average_precision
                for label_stats in self.label_statistics_dict.values()
                if label_stats.num_annotations > 0
            ]
        )

    def average_precision_for_label(self, label):
        return self.label_statistics_dict[label].average_precision

    def __call__(
        self,
        predicted_boxes,
        predicted_labels,
        predicted_scores,
        target_boxes,
        target_labels,
    ):
        for (
            predicted_boxes_for_sample,
            predicted_labels_for_sample,
            predicted_scores_for_sample,
            target_boxes_for_sample,
            target_labels_for_sample,
        ) in zip(
            predicted_boxes,
            predicted_labels,
            predicted_scores,
            target_boxes,
            target_labels,
        ):
            for label in self.label_statistics_dict.keys():
                (
                    predicted_boxes_for_label,
                    _,
                    scores_for_predicted_label,
                    target_boxes_for_label,
                    _,
                ) = self._create_predicted_and_targets_given_label(
                    label,
                    predicted_boxes_for_sample,
                    predicted_labels_for_sample,
                    predicted_scores_for_sample,
                    target_boxes_for_sample,
                    target_labels_for_sample,
                )

                self._update_statistics_for_label(
                    label,
                    predicted_boxes_for_label,
                    scores_for_predicted_label,
                    target_boxes_for_label,
                )

    def _create_predicted_and_targets_given_label(
        self,
        label,
        predicted_boxes,
        predicted_labels,
        predicted_scores,
        target_boxes,
        target_labels,
    ):
        idxs_for_label_predicted = np.where(predicted_labels == label)[0]
        predicted_boxes_for_label = predicted_boxes[idxs_for_label_predicted]
        predicted_labels_for_label = predicted_labels[idxs_for_label_predicted]
        predicted_scores_for_label = predicted_scores[idxs_for_label_predicted]

        idxs_for_label_target = np.where(target_labels == label)[0]
        target_boxes_for_label = target_boxes[idxs_for_label_target]
        target_labels_for_label = target_labels[idxs_for_label_target]
        return (
            predicted_boxes_for_label,
            predicted_labels_for_label,
            predicted_scores_for_label,
            target_boxes_for_label,
            target_labels_for_label,
        )

    def _update_statistics_for_label(
        self,
        label,
        predicted_boxes_for_label,
        scores_for_predicted_label,
        target_boxes_for_label,
    ):

        if len(predicted_boxes_for_label) == 0 and len(target_boxes_for_label) == 0:
            return

        self.label_statistics_dict[label].num_annotations += len(target_boxes_for_label)

        detected_annotations = []

        if self.nms_threshold is not None:
            idxs_to_keep = self._nms(
                predicted_boxes_for_label,
                scores_for_predicted_label,
                self.nms_threshold,
            )
            predicted_boxes_for_label = predicted_boxes_for_label[idxs_to_keep]
            scores_for_predicted_label = scores_for_predicted_label[idxs_to_keep]

        for predicted_box_for_label, score in zip(
            predicted_boxes_for_label, scores_for_predicted_label
        ):

            self.label_statistics_dict[label].scores.append(score)

            # If there are no annotations, but a box was predicted anyways, we found a false positive
            if len(target_boxes_for_label) == 0:
                self.label_statistics_dict[label].false_positives.append(1)
                self.label_statistics_dict[label].true_positives.append(0)
                continue

            # Find the predicted box that has the highest IoU with all the annotations
            overlaps = self._compute_overlap(
                np.expand_dims(predicted_box_for_label, axis=0), target_boxes_for_label
            )
            assigned_annotation = np.argmax(overlaps, axis=1)
            max_overlap = overlaps[0, assigned_annotation]

            # If the IoU of the detected box is greater than the threshold & it hasn't been used, we found a true position
            if (
                max_overlap >= self.iou_threshold
                and assigned_annotation not in detected_annotations
            ):
                self.label_statistics_dict[label].false_positives.append(0)
                self.label_statistics_dict[label].true_positives.append(1)
                detected_annotations.append(assigned_annotation)
            # Or else we found a false_positive
            else:
                self.label_statistics_dict[label].false_positives.append(1)
                self.label_statistics_dict[label].true_positives.append(0)

    # --------------------------------------------------------
    # Fast R-CNN
    # Copyright (c) 2015 Microsoft
    # Licensed under The MIT License [see LICENSE for details]
    # Written by Ross Girshick
    # --------------------------------------------------------
    def _nms(self, boxes, scores, threshold):
        """
        boxes is a numpy array: num_boxes, 4
        scores is a  numpy array: num_boxes
        """
        x1 = boxes[:, 0]
        y1 = boxes[:, 1]
        x2 = boxes[:, 2]
        y2 = boxes[:, 3]

        areas = (x2 - x1 + 1) * (y2 - y1 + 1)
        order = scores.argsort()[::-1]  # get boxes with more ious first

        keep = []
        while order.size > 0:
            i = order[0]  # pick maxmum iou box
            keep.append(i)
            xx1 = np.maximum(x1[i], x1[order[1:]])
            yy1 = np.maximum(y1[i], y1[order[1:]])
            xx2 = np.minimum(x2[i], x2[order[1:]])
            yy2 = np.minimum(y2[i], y2[order[1:]])

            w = np.maximum(0.0, xx2 - xx1 + 1)  # maximum width
            h = np.maximum(0.0, yy2 - yy1 + 1)  # maxiumum height
            inter = w * h
            ovr = inter / (areas[i] + areas[order[1:]] - inter)

            inds = np.where(ovr <= threshold)[0]
            order = order[inds + 1]

        return keep

    def _compute_overlap(self, boxes, query_boxes):
        """
        Args
            a: (N, 4) ndarray of float
            b: (K, 4) ndarray of float
        Returns
            overlaps: (N, K) ndarray of overlap between boxes and query_boxes
        """
        N = boxes.shape[0]
        K = query_boxes.shape[0]
        overlaps = np.zeros((N, K), dtype=np.float64)
        for k in range(K):
            box_area = (query_boxes[k, 2] - query_boxes[k, 0]) * (
                query_boxes[k, 3] - query_boxes[k, 1]
            )
            for n in range(N):
                iw = min(boxes[n, 2], query_boxes[k, 2]) - max(
                    boxes[n, 0], query_boxes[k, 0]
                )
                if iw > 0:
                    ih = min(boxes[n, 3], query_boxes[k, 3]) - max(
                        boxes[n, 1], query_boxes[k, 1]
                    )
                    if ih > 0:
                        ua = np.float64(
                            (boxes[n, 2] - boxes[n, 0]) * (boxes[n, 3] - boxes[n, 1])
                            + box_area
                            - iw * ih
                        )
                        overlaps[n, k] = iw * ih / ua
        return overlaps


lookup = {
    "MeanAveragePrecision": MeanAveragePrecision,
    "mean_average_precision_torchvision_fasterrcnn_transform_fn": mean_average_precision_torchvision_fasterrcnn_transform_fn,
}



from .config_loader import dump_config_to_path
from .constants import VARIABLE_KEY


class Configuration:
    def __init__(self, raw_config):
        self.raw_config = raw_config

    @property
    def variables(self):
        return self.raw_config[VARIABLE_KEY]

    @property
    def model_save_path(self):
        return self.raw_config[VARIABLE_KEY]["model_save_path"]

    @model_save_path.setter
    def model_save_path(self, path):
        self.raw_config[VARIABLE_KEY]["model_save_path"] = str(path)

    @property
    def root_save_directory(self):
        return self.raw_config[VARIABLE_KEY]["root_save_directory"]

    @root_save_directory.setter
    def root_save_directory(self, path):
        self.raw_config[VARIABLE_KEY]["root_save_directory"] = str(path)

    def dump_config_to_path(self, path):
        dump_config_to_path(self.raw_config, path)



from .callback import *
from .callback import Callback
from .handler import *
from .handler import CallbackHandler
from .recorders import *
from .recorders import lookup as recorder_lookup
from .savers import *
from .savers import lookup as saver_lookup

lookup = {
    **recorder_lookup,
    **saver_lookup,
    "Callback": Callback,
    "CallbackHandler": CallbackHandler,
}



from collections import defaultdict

import numpy as np


def filter_image_bboxes_samples_by_labels(samples, labels_to_keep=None):
    if labels_to_keep is None:
        return samples

    samples_lookup_by_label = create_samples_lookup_by_label(samples)

    filtered_samples = {}
    for label in labels_to_keep:
        for sample in samples_lookup_by_label[label]:
            filtered_samples[id(sample)] = sample
    return np.array([s for s in filtered_samples.values()])


def create_samples_lookup_by_label(samples):
    samples_lookup = defaultdict(list)
    for sample in samples:
        unique_labels = list(set(sample.labels.data))
        for label in unique_labels:
            samples_lookup[label].append(sample)
    return samples_lookup


lookup = {
    "filter_image_bboxes_samples_by_labels": filter_image_bboxes_samples_by_labels,
    "create_samples_lookup_by_label": create_samples_lookup_by_label,
}



from abc import ABC, abstractmethod


class MapDictionaryToMethodArguments(ABC):
    @abstractmethod
    def to_call_kwargs(self):
        return


class TransformToCallKeywordArguments(ABC):
    @abstractmethod
    def transform_to_kwargs(self):
        return


class MissingMethodArgumentException(Exception):
    pass


lookup = {
    "MapDictionaryToMethodArguments": MapDictionaryToMethodArguments,
    "MissingMethodArgumentException": MissingMethodArgumentException,
}



from torch.optim.lr_scheduler import CosineAnnealingLR, _LRScheduler


class DelayerScheduler(_LRScheduler):
    """Starts with a flat lr schedule until it reaches N epochs the applies a scheduler
    Args:
            optimizer (Optimizer): Wrapped optimizer.
            delay_epochs: number of epochs to keep the initial lr until starting aplying the scheduler
            after_scheduler: after target_epoch, use this scheduler(eg. ReduceLROnPlateau)
    """

    def __init__(self, optimizer, delay_epochs, after_scheduler):
        self.delay_epochs = delay_epochs
        self.after_scheduler = after_scheduler
        self.finished = False
        super().__init__(optimizer)

    def get_lr(self):
        if self.last_epoch >= self.delay_epochs:
            if not self.finished:
                self.after_scheduler.base_lrs = self.base_lrs
                self.finished = True
            return self.after_scheduler.get_lr()

        return self.base_lrs

    def step(self, epoch=None):
        if self.finished:
            if epoch is None:
                self.after_scheduler.step(None)
            else:
                self.after_scheduler.step(epoch - self.delay_epochs)
        else:
            return super(DelayerScheduler, self).step(epoch)


def DelayedCosineAnnealingLR(optimizer, delay_epochs, cosine_annealing_epochs):
    base_scheduler = CosineAnnealingLR(optimizer, cosine_annealing_epochs)
    return DelayerScheduler(optimizer, delay_epochs, base_scheduler)



from .model_checkpoint import *
from .model_checkpoint import lookup as model_checkpoint_lookup
from .on_epoch_end_transformer import *
from .on_epoch_end_transformer import lookup as on_epoch_end_transformer_lookup

lookup = {**model_checkpoint_lookup, **on_epoch_end_transformer_lookup}



import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision

from ....utils.model import model_sizes
from ..utils.layers import conv2d, conv2d_relu, conv2d_trans


class UnetBlock(nn.Module):
    "A basic U-Net block."

    def __init__(self, up_in_c, x_in_c, hook):
        super().__init__()
        self.hook = hook
        ni = up_in_c
        self.upconv = conv2d_trans(ni, ni // 2)  # H, W -> 2H, 2W
        ni = ni // 2 + x_in_c
        self.conv1 = conv2d(ni, ni // 2)
        ni = ni // 2
        self.conv2 = conv2d(ni, ni)
        self.bn = nn.BatchNorm2d(ni)

    def forward(self, up_in):
        up_out = self.upconv(up_in)

        try:
            cat_x = torch.cat([up_out, self.hook.stored], dim=1)
        except Exception as e:
            import pdb

            pdb.set_trace()
            print(e)
            pass
        x = F.relu(self.conv1(cat_x))
        x = F.relu(self.conv2(x))
        return self.bn(x)


class DynamicUnet(nn.Sequential):
    "Create a U-Net from a given architecture."

    def __init__(self, encoder, n_classes):
        imsize = (256, 256)
        sfs_szs, x, self.sfs = model_sizes(encoder, size=imsize)
        sfs_idxs = reversed(_get_sfs_idxs(sfs_szs))

        ni = sfs_szs[-1][1]
        middle_conv = nn.Sequential(
            conv2d_relu(ni, ni * 2, bn=True), conv2d_relu(ni * 2, ni, bn=True)
        )
        x = middle_conv(x)
        layers = [encoder, nn.ReLU(), middle_conv]

        for idx in sfs_idxs:
            up_in_c, x_in_c = int(x.shape[1]), int(sfs_szs[idx][1])
            unet_block = UnetBlock(up_in_c, x_in_c, self.sfs[idx])
            layers.append(unet_block)
            x = unet_block(x)

        ni = unet_block.conv2.out_channels
        if imsize != sfs_szs[0][-2:]:
            layers.append(conv2d_trans(ni, ni))
        layers.append(conv2d(ni, n_classes, 1))
        super().__init__(*layers)

    def __del__(self):
        if hasattr(self, "sfs"):
            self.sfs.remove()


def _get_sfs_idxs(sizes):
    "Get the indexes of the layers where the size of the activation changes."
    feature_szs = [size[-1] for size in sizes]
    sfs_idxs = list(
        np.where(np.array(feature_szs[:-1]) != np.array(feature_szs[1:]))[0]
    )
    if feature_szs[0] != feature_szs[1]:
        sfs_idxs = [0] + sfs_idxs
    return sfs_idxs


def create_encoder(name, pretrained=True):
    if name == "pytorch_resnet18":
        return nn.Sequential(
            *list(torchvision.models.resnet.resnet18(pretrained=pretrained).children())[
                :-2
            ]
        )
    elif name == "pytorch_resnet34":
        return nn.Sequential(
            *list(torchvision.models.resnet.resnet34(pretrained=pretrained).children())[
                :-2
            ]
        )
    elif name == "pytorch_resnet50":
        return nn.Sequential(
            *list(torchvision.models.resnet.resnet50(pretrained=pretrained).children())[
                :-2
            ]
        )
    elif name == "pytorch_resnet101":
        return nn.Sequential(
            *list(
                torchvision.models.resnet.resnet101(pretrained=pretrained).children()
            )[:-2]
        )
    elif name == "pytorch_resnet152":
        return nn.Sequential(
            *list(
                torchvision.models.resnet.resnet152(pretrained=pretrained).children()
            )[:-2]
        )
    else:
        print(f"A method to extract the encoder from {name} is not defined")


def unet_with_resnet18_backbone(n_classes=1, pretrained=True, include_encoder=False):
    encoder = create_encoder("pytorch_resnet18", pretrained=pretrained)
    unet = DynamicUnet(encoder, n_classes=n_classes)
    if include_encoder:
        return encoder, unet
    return unet


def unet_with_resnet34_backbone(n_classes=1, pretrained=True, include_encoder=False):
    encoder = create_encoder("pytorch_resnet34", pretrained=pretrained)
    unet = DynamicUnet(encoder, n_classes=n_classes)
    if include_encoder:
        return encoder, unet
    return unet


def unet_with_resnet50_backbone(n_classes=1, pretrained=True, include_encoder=False):
    encoder = create_encoder("pytorch_resnet50", pretrained=pretrained)
    unet = DynamicUnet(encoder, n_classes=n_classes)
    if include_encoder:
        return encoder, unet
    return unet


def unet_with_resnet101_backbone(n_classes=1, pretrained=True, include_encoder=False):
    encoder = create_encoder("pytorch_resnet101", pretrained=pretrained)
    unet = DynamicUnet(encoder, n_classes=n_classes)
    if include_encoder:
        return encoder, unet
    return unet


def unet_with_resnet152_backbone(n_classes=1, pretrained=True, include_encoder=False):
    encoder = create_encoder("pytorch_resnet152", pretrained=pretrained)
    unet = DynamicUnet(encoder, n_classes=n_classes)
    if include_encoder:
        return encoder, unet
    return unet



import numpy as np

from ......vision.utils import imread_rgb
from .sample import ImageClassificationSample


def create_image_classification_sample(
    image_data=None,
    image_path=None,
    load_image_fn=imread_rgb,
    label_data=None,
    label_path=None,
    load_label_fn=np.load,
    name=None,
):
    return ImageClassificationSample.create(
        image_data=image_data,
        image_path=image_path,
        load_image_fn=load_image_fn,
        label_data=label_data,
        label_path=label_path,
        load_label_fn=load_label_fn,
        name=name,
    )


lookup = {"create_image_classification_sample": create_image_classification_sample}



import traceback
from copy import deepcopy

import networkx as nx

from ..logger import logger
from .config import Configuration
from .config_loader import (
    dump_config_to_string,
    load_config_from_path,
    load_config_from_paths,
    load_config_from_string,
    load_config_from_strings,
    replace_config_variables,
)
from .constants import RESOURCE_KEY
from .graph_construction import Reference, find_references, flatten_resources_dict, replace_arguments, replace_should_remove
from .node import Node, NodeRemoveMode, remove_node_arguments


class Pipeline:
    def __init__(self, graph, config, lookup):
        self.graph = graph
        self.config = config
        self.lookup = lookup

    @classmethod
    def create_from_graph_with_no_config_and_lookup(cls, graph):
        graph = cls._add_edges_to_graph(graph)
        return cls(graph, None, None)

    @classmethod
    def create_from_config(cls, config, lookup):
        assert (
            config.get(RESOURCE_KEY) is not None
        ), f"There is no {RESOURCE_KEY} key in the configuration file"

        raw_config = load_config_from_string(dump_config_to_string(config))

        config = Configuration(deepcopy(raw_config))

        variable_replaced_config = replace_config_variables(deepcopy(raw_config))
        flattened_resources = flatten_resources_dict(
            variable_replaced_config[RESOURCE_KEY]
        )

        graph = nx.DiGraph()
        graph = cls._add_nodes_to_graph(graph, flattened_resources, lookup)
        graph = cls._add_edges_to_graph(graph)
        return cls(graph, config, lookup)

    @classmethod
    def create_from_config_string(
        cls, config_string, lookup, with_variable_replacement=False
    ):
        config = load_config_from_string(
            config_string, with_variable_replacement=with_variable_replacement
        )
        return cls.create_from_config(config, lookup)

    @classmethod
    def create_from_config_strings(
        cls, config_strings, lookup, with_variable_replacement=False
    ):
        config = load_config_from_strings(
            config_strings, with_variable_replacement=with_variable_replacement
        )
        return cls.create_from_config(config, lookup)

    @classmethod
    def create_from_config_path(
        cls, config_path, lookup, with_variable_replacement=False
    ):
        config = load_config_from_path(
            config_path, with_variable_replacement=with_variable_replacement
        )
        return cls.create_from_config(config, lookup)

    @classmethod
    def create_from_config_paths(
        cls, config_paths, lookup, with_variable_replacement=False
    ):
        config = load_config_from_paths(
            config_paths, with_variable_replacement=with_variable_replacement
        )
        return cls.create_from_config(config, lookup)

    @staticmethod
    def _add_nodes_to_graph(graph, resources, lookup):
        logger.info("Adding nodes to graph")
        for name, resource in resources.items():
            assert (
                "properties" in resource
            ), f"The properties key isn't defined for node: {name}"
            references = find_references(resource)
            properties = Node.load_properties_with_default_values(
                resource["properties"], lookup
            )
            node = Node(name=name, references=references, **properties)
            graph.add_node(name, node=node)
        return graph

    @staticmethod
    def _add_edges_to_graph(graph):
        logger.info("Adding edges to graph")
        for name, node_wrapper in graph.nodes(data=True):
            node = node_wrapper["node"]
            for reference in node.references:
                referenced_node_name = reference.ref_node_name

                assert (
                    referenced_node_name in graph.nodes
                ), f"The reference {referenced_node_name} in node {node.name} does not exist, \
                    try checking if {referenced_node_name} is defined in the configuration"

                referenced_node = graph.nodes(data=True)[referenced_node_name]["node"]
                referenced_node.referenced_by.append(node)

                graph.add_edge(referenced_node_name, name)
        return graph

    @property
    def variables(self):
        return self.config.variables

    @property
    def sorted_node_names(self):
        return list(nx.algorithms.dag.topological_sort(self.graph))

    def get_node(self, name):
        return self.graph.nodes(data=True)[name]["node"]

    def get_node_output_lookup(self, name):
        return self.get_node(name).output

    def get_node_output(self, name, output_lookup_key=None):
        output_lookup = self.get_node_output_lookup(name)
        if output_lookup is None:
            return None

        if output_lookup_key is None and len(output_lookup) == 1:
            output = list(output_lookup.values())[0]
            return output
        else:
            return output_lookup[output_lookup_key]

    def node_exists(self, name):
        return name in self.graph.nodes()

    def _nodes_to_be_removed(self, tags_to_be_removed=[]):
        nodes_to_be_removed = []
        for node_name in self.graph:
            node = self.graph.nodes(data=True)[node_name]["node"]
            node_tag_is_in_tags_to_be_removed = (
                len(set(node.tags) & set(tags_to_be_removed)) > 0
            )
            if node.should_remove or node_tag_is_in_tags_to_be_removed:
                nodes_to_be_removed.append(node)
        return nodes_to_be_removed

    def run(self, to_node=None, remove_nodes_with_tags=[], fail_early=False):
        has_error = False
        errors = []
        for node_name in nx.algorithms.dag.topological_sort(self.graph):
            try:
                node = self.graph.nodes(data=True)[node_name]["node"]

                # Replace all `References` of a removed node in the arguments
                # of the current `node` to SHOULD_REMOVE_ARGUMENT.
                # This is needed # as nodes that are removed from the graph won't
                # be run, as a result there will be no output and references to a
                # node's output for a node that hasn't been run will fail
                self._update_node_arguments_for_removed_nodes(
                    node, self._nodes_to_be_removed(remove_nodes_with_tags)
                )

                # replace all references in the argument of the current `node`
                self._update_node_arguments_references(node)

                # update `node.should_run` & replace the reference if it exists
                # self._update_node_should_run(node)

                # update `node.should_remove` & replace the reference if it exists
                self._update_node_should_remove(node, remove_nodes_with_tags)

                logger.info(
                    f"Processing node: {node_name} | Should run: {not node.should_remove}"
                )

                if not node.should_remove:
                    logger.debug(f"Running node: {node.name}")
                    node.create_output()

                if to_node == node_name:
                    break

            except Exception as e:
                if fail_early:
                    raise e
                has_error = True
                error = "\n".join(
                    [
                        f"An error has happened when trying to run node: {node_name}",
                        f"The state when the node failed is: {node}",
                        f"{traceback.format_exc()}",
                        "\n",
                    ]
                )
                errors.append(error)

        if has_error:
            raise ValueError(
                f"There were errors while running the graph:\n\n{(''.join(errors))}"
            )

    def _update_node_arguments_for_removed_nodes(self, node, nodes_to_be_removed):
        logger.debug(f"Updating node arguments for removed nodes: {node.name}")
        node.arguments = remove_node_arguments(node.arguments, nodes_to_be_removed)

    def _update_node_arguments_references(self, node):
        logger.debug(f"Replacing references for node: {node.name}")
        logger.debug(f"Original arguments: {node.arguments}")
        reference_replaced_arguments = replace_arguments(self, node)
        node.reference_replaced_arguments.update(**reference_replaced_arguments)
        logger.debug(f"Replaced arguments: {node.reference_replaced_arguments}")

    def _update_node_should_remove(self, node, tags_to_be_removed):
        logger.debug(f"Updating should_remove property for node: {node.name}")

        # case for NodeRemoveMode.ONLY
        node_should_remove = self._get_node_should_remove(node, tags_to_be_removed)
        node.should_remove = node_should_remove

        if node.remove_mode == NodeRemoveMode.PROPAGATE:
            for referenced_node in node.referenced_by:
                if (
                    isinstance(referenced_node.should_remove, Reference)
                    or referenced_node.should_remove is True
                ):
                    continue
                else:
                    referenced_node.should_remove = node_should_remove
                    self._update_node_should_remove(referenced_node, tags_to_be_removed)

        elif node.remove_mode == NodeRemoveMode.AUTO:
            for referenced_node in node.referenced_by:
                if (
                    isinstance(referenced_node.should_remove, Reference)
                    or referenced_node.should_remove is True
                ):
                    continue
                else:
                    for arg_name, arg_value in referenced_node.arguments.items():
                        # if the reference node uses the current node & there is no default argument for it, then propagate
                        if (
                            isinstance(arg_value, Reference)
                            and arg_value.ref_node_name == node.name
                        ):
                            logger.debug(
                                f"Remove argument '{arg_name}' with value '{arg_value}' for node {node.name}: True"
                            )
                            referenced_node.should_remove = node_should_remove
                            self._update_node_should_remove(
                                referenced_node, tags_to_be_removed
                            )
                        else:
                            logger.debug(
                                f"Remove argument '{arg_name}' with value '{arg_value}' for node {node.name}: False"
                            )

    def _get_node_should_remove(self, node, tags_to_be_removed):
        if isinstance(node.should_remove, Reference):
            node_should_remove_property = replace_should_remove(self.graph, node)
        else:
            node_should_remove_property = node.should_remove
        node_tag_is_in_tags_to_be_removed = (
            len(set(node.tags) & set(tags_to_be_removed)) > 0
        )
        return node_should_remove_property or node_tag_is_in_tags_to_be_removed

    def _run_node(self, node):
        node.create_output()



from abc import ABC
from enum import Enum, IntEnum

from ...utils.convenience import debug_log_on_call


class CallbackHandler:
    @debug_log_on_call
    def __init__(self, callbacks):
        self.context = CallbackHandlerContext()
        self.callbacks = callbacks
        self.callbacks = sorted(self.callbacks, key=lambda cb: getattr(cb, "_order", 0))

    def get_callback_by_name(self, name):
        for cb in self.callbacks:
            if cb.name == name:
                return cb

        for cb in self.callbacks:
            if name in cb.name:
                return cb

        return None

    def set_context_value(self, key, value):
        self.context[key] = value

    def get_context_value(self, key):
        return self.context[key]

    @debug_log_on_call
    def on_setup(self):
        for cb in self.callbacks:
            cb.on_setup(self.context)

    @debug_log_on_call
    def on_init_begin(self):
        for cb in self.callbacks:
            cb.on_init_begin(self.context)

    @debug_log_on_call
    def on_init_end(self):
        for cb in self.callbacks:
            cb.on_init_end(self.context)

    @debug_log_on_call
    def on_train_begin(self):
        for cb in self.callbacks:
            cb.on_train_begin(self.context)

    @debug_log_on_call
    def on_train_epoch_begin(self):
        self.context.current_phase = Phase.TRAIN
        for cb in self.callbacks:
            cb.on_train_epoch_begin(self.context)

    @debug_log_on_call
    def on_train_batch_begin(self):
        for cb in self.callbacks:
            cb.on_train_batch_begin(self.context)
        return self.context.train_batch

    @debug_log_on_call
    def on_train_loss_begin(self):
        for cb in self.callbacks:
            cb.on_train_loss_begin(self.context)
        return self.context.train_prediction

    @debug_log_on_call
    def on_train_step_end(self):
        for cb in self.callbacks:
            cb.on_train_step_end(self.context)
        return self.context.train_output

    @debug_log_on_call
    def on_backward_begin(self):
        for cb in self.callbacks:
            cb.on_backward_begin(self.context)
        return self.context.train_loss

    @debug_log_on_call
    def on_backward_end(self):
        for cb in self.callbacks:
            cb.on_backward_end(self.context)

    @debug_log_on_call
    def on_train_batch_end(self):
        for cb in self.callbacks:
            cb.on_train_batch_end(self.context)

    @debug_log_on_call
    def on_validation_begin(self):
        for cb in self.callbacks:
            cb.on_validation_begin(self.context)

    @debug_log_on_call
    def on_validation_epoch_begin(self):
        self.context.current_phase = Phase.VALIDATION
        for cb in self.callbacks:
            cb.on_validation_epoch_begin(self.context)

    @debug_log_on_call
    def on_validation_batch_begin(self):
        for cb in self.callbacks:
            cb.on_validation_batch_begin(self.context)
        return self.context.validation_batch

    @debug_log_on_call
    def on_validation_loss_begin(self):
        for cb in self.callbacks:
            cb.on_validation_loss_begin(self.context)
        return self.context.validation_prediction

    @debug_log_on_call
    def on_validation_step_end(self):
        for cb in self.callbacks:
            cb.on_validation_step_end(self.context)
        return self.context.validation_output

    @debug_log_on_call
    def on_validation_batch_end(self):
        for cb in self.callbacks:
            cb.on_validation_batch_end(self.context)

    @debug_log_on_call
    def on_validation_epoch_end(self):
        for cb in self.callbacks:
            cb.on_validation_epoch_end(self.context)
        return self.context.all_validation_output

    @debug_log_on_call
    def on_validation_end(self):
        for cb in self.callbacks:
            cb.on_validation_end(self.context)

    @debug_log_on_call
    def on_train_epoch_end(self):
        for cb in self.callbacks:
            cb.on_train_epoch_end(self.context)
        return self.context.all_train_output

    @debug_log_on_call
    def on_train_end(self):
        for cb in self.callbacks:
            cb.on_train_end(self.context)

    @debug_log_on_call
    def on_test_begin(self):
        for cb in self.callbacks:
            cb.on_test_begin(self.context)

    @debug_log_on_call
    def on_test_epoch_begin(self):
        self.context.current_phase = Phase.TEST
        for cb in self.callbacks:
            cb.on_test_epoch_begin(self.context)

    @debug_log_on_call
    def on_test_batch_begin(self):
        for cb in self.callbacks:
            cb.on_test_batch_begin(self.context)
        return self.context.test_batch

    @debug_log_on_call
    def on_test_loss_begin(self):
        for cb in self.callbacks:
            cb.on_test_loss_begin(self.context)
        return self.context.test_prediction

    @debug_log_on_call
    def on_test_step_end(self):
        for cb in self.callbacks:
            cb.on_test_step_end(self.context)
        return self.context.test_output

    @debug_log_on_call
    def on_test_batch_end(self):
        for cb in self.callbacks:
            cb.on_test_batch_end(self.context)

    @debug_log_on_call
    def on_test_epoch_end(self):
        for cb in self.callbacks:
            cb.on_test_epoch_end(self.context)
        return self.context.all_test_output

    @debug_log_on_call
    def on_test_end(self):
        for cb in self.callbacks:
            cb.on_test_end(self.context)

    @debug_log_on_call
    def on_teardown(self):
        for cb in self.callbacks:
            cb.on_teardown(self.context)


Phase = IntEnum("Phase", "TRAIN VALIDATION TEST")
CallbackHandlerContextValues = IntEnum(
    "CallbackHandlerContextValues",
    "SAVE_DIRECTORY LOSS MODEL TRAINER TRAIN_BATCH TRAIN_PREDICTION TRAIN_LOSS TRAIN_METRIC TRAIN_OUTPUT ALL_TRAIN_OUTPUT \
    VALIDATION_BATCH VALIDATION_PREDICTION VALIDATION_LOSS VALIDATION_METRIC VALIDATION_OUTPUT ALL_VALIDATION_OUTPUT \
    TEST_BATCH TEST_PREDICTION TEST_LOSS TEST_METRIC TEST_OUTPUT ALL_TEST_OUTPUT CURRENT_EPOCH CURRENT_PHASE CURRENT_STEP",
)


class CallbackHandlerContext(dict):
    def __init__(self):
        self.current_epoch = 0

    @property
    def save_directory(self):
        return self.get(CallbackHandlerContextValues.SAVE_DIRECTORY)

    @save_directory.setter
    def save_directory(self, value):
        self[CallbackHandlerContextValues.SAVE_DIRECTORY] = value

    @property
    def loss(self):
        return self.get(CallbackHandlerContextValues.LOSS)

    @property
    def model(self):
        return self.get(CallbackHandlerContextValues.MODEL)

    @property
    def trainer(self):
        return self.get(CallbackHandlerContextValues.TRAINER)

    @property
    def current_phase(self):
        return self.get(CallbackHandlerContextValues.CURRENT_PHASE)

    @current_phase.setter
    def current_phase(self, value):
        self[CallbackHandlerContextValues.CURRENT_PHASE] = value

    @property
    def current_epoch(self):
        return self.get(CallbackHandlerContextValues.CURRENT_EPOCH)

    @current_epoch.setter
    def current_epoch(self, value):
        self[CallbackHandlerContextValues.CURRENT_EPOCH] = value

    @property
    def current_step(self):
        return self.get(CallbackHandlerContextValues.CURRENT_STEP)

    @current_step.setter
    def current_step(self, value):
        self[CallbackHandlerContextValues.CURRENT_STEP] = value

    @property
    def train_batch(self):
        return self.get(CallbackHandlerContextValues.TRAIN_BATCH)

    @train_batch.setter
    def train_batch(self, value):
        self[CallbackHandlerContextValues.TRAIN_BATCH] = value

    @property
    def train_prediction(self):
        return self.get(CallbackHandlerContextValues.TRAIN_PREDICTION)

    @train_prediction.setter
    def train_prediction(self, value):
        self[CallbackHandlerContextValues.TRAIN_PREDICTION] = value

    @property
    def train_loss(self):
        return self.get(CallbackHandlerContextValues.TRAIN_LOSS)

    @train_loss.setter
    def train_loss(self, value):
        self[CallbackHandlerContextValues.TRAIN_LOSS] = value

    @property
    def train_metric(self):
        return self.get(CallbackHandlerContextValues.TRAIN_METRIC)

    @train_metric.setter
    def train_metric(self, value):
        self[CallbackHandlerContextValues.TRAIN_METRIC] = value

    @property
    def train_output(self):
        return self.get(CallbackHandlerContextValues.TRAIN_OUTPUT)

    @train_output.setter
    def train_output(self, value):
        self[CallbackHandlerContextValues.TRAIN_OUTPUT] = value

    @property
    def all_train_output(self):
        return self.get(CallbackHandlerContextValues.ALL_TRAIN_OUTPUT)

    @all_train_output.setter
    def all_train_output(self, value):
        self[CallbackHandlerContextValues.ALL_TRAIN_OUTPUT] = value

    @property
    def validation_batch(self):
        return self.get(CallbackHandlerContextValues.VALIDATION_BATCH)

    @validation_batch.setter
    def validation_batch(self, value):
        self[CallbackHandlerContextValues.VALIDATION_BATCH] = value

    @property
    def validation_prediction(self):
        return self.get(CallbackHandlerContextValues.VALIDATION_PREDICTION)

    @validation_prediction.setter
    def validation_prediction(self, value):
        self[CallbackHandlerContextValues.VALIDATION_PREDICTION] = value

    @property
    def validation_loss(self):
        return self.get(CallbackHandlerContextValues.VALIDATION_LOSS)

    @validation_loss.setter
    def validation_loss(self, value):
        self[CallbackHandlerContextValues.VALIDATION_LOSS] = value

    @property
    def validation_metric(self):
        return self.get(CallbackHandlerContextValues.VALIDATION_METRIC)

    @validation_metric.setter
    def validation_metric(self, value):
        self[CallbackHandlerContextValues.VALIDATION_METRIC] = value

    @property
    def validation_output(self):
        return self.get(CallbackHandlerContextValues.VALIDATION_OUTPUT)

    @validation_output.setter
    def validation_output(self, value):
        self[CallbackHandlerContextValues.VALIDATION_OUTPUT] = value

    @property
    def all_validation_output(self):
        return self.get(CallbackHandlerContextValues.ALL_VALIDATION_OUTPUT)

    @all_validation_output.setter
    def all_validation_output(self, value):
        self[CallbackHandlerContextValues.ALL_VALIDATION_OUTPUT] = value

    @property
    def test_batch(self):
        return self.get(CallbackHandlerContextValues.TEST_BATCH)

    @test_batch.setter
    def test_batch(self, value):
        self[CallbackHandlerContextValues.TEST_BATCH] = value

    @property
    def test_prediction(self):
        return self.get(CallbackHandlerContextValues.TEST_PREDICTION)

    @test_prediction.setter
    def test_prediction(self, value):
        self[CallbackHandlerContextValues.TEST_PREDICTION] = value

    @property
    def test_loss(self):
        return self.get(CallbackHandlerContextValues.TEST_LOSS)

    @test_loss.setter
    def test_loss(self, value):
        self[CallbackHandlerContextValues.TEST_LOSS] = value

    @property
    def test_metric(self):
        return self.get(CallbackHandlerContextValues.TEST_METRIC)

    @test_metric.setter
    def test_metric(self, value):
        self[CallbackHandlerContextValues.TEST_METRIC] = value

    @property
    def test_output(self):
        return self.get(CallbackHandlerContextValues.TEST_OUTPUT)

    @test_output.setter
    def test_output(self, value):
        self[CallbackHandlerContextValues.TEST_OUTPUT] = value

    @property
    def all_test_output(self):
        return self.get(CallbackHandlerContextValues.ALL_TEST_OUTPUT)

    @all_test_output.setter
    def all_test_output(self, value):
        self[CallbackHandlerContextValues.ALL_TEST_OUTPUT] = value



import pytorch_lightning as pl

from roheboam.engine.utils.convenience import timing

from ....core.callbacks import CallbackHandlerContextValues
from ....utils.convenience import debug_log_on_call


class LightningModuleWrapper:
    def __init__(self, lightning_module):
        self.lightning_module = lightning_module

    def __getattr__(self, name):
        return getattr(self.lightning_module, name)

    def __setattr__(self, name, value):
        if name == "lightning_module":
            super().__setattr__(name, value)
        else:
            setattr(self.lightning_module, name, value)


class PytorchLightningModuleHooksToCallbackHandlerAdapter(pl.LightningModule):
    """
    This wraps the LightningModule use the `callback_handler` to handle both the trainer hooks and module hooks
    """

    @debug_log_on_call
    def __init__(self, lightning_module, callback_handler):
        lightning_module.callback_handler = callback_handler

        self.callback_handler = callback_handler
        self.lightning_module = LightningModuleWrapper(lightning_module)
        self.callback_handler.set_context_value(
            CallbackHandlerContextValues.LOSS, self.lightning_module.loss
        )

    def __setattr__(self, name, value):
        if name in ["callback_handler", "lightning_module"]:
            super().__setattr__(name, value)
        else:
            setattr(self.lightning_module, name, value)

    def __getattr__(self, name):
        if name in ["callback_handler", "lightning_module"]:
            return super().__getattr__(name)
        else:
            return getattr(self.lightning_module, name)

    @debug_log_on_call
    def prepare_data(self):
        self.lightning_module.prepare_data()

    @debug_log_on_call
    def train_dataloader(self):
        return self.lightning_module.train_dataloader()

    @debug_log_on_call
    def val_dataloader(self):
        return self.lightning_module.val_dataloader()

    @debug_log_on_call
    def test_dataloader(self):
        return self.lightning_module.test_dataloader()

    @debug_log_on_call
    def configure_optimizers(self):
        return self.lightning_module.configure_optimizers()

    @debug_log_on_call
    def training_step(self, batch, batch_idx):
        self.callback_handler.set_context_value(
            CallbackHandlerContextValues.TRAIN_BATCH, batch
        )
        batch = self.callback_handler.on_train_batch_begin()
        loss = self.lightning_module.training_step(batch, batch_idx)
        self.callback_handler.set_context_value(
            CallbackHandlerContextValues.TRAIN_LOSS, loss.get("loss")
        )
        return loss

    @debug_log_on_call
    def forward(self, x):
        return self.lightning_module.forward(x)

    @debug_log_on_call
    def training_step_end(self, outputs):
        module_outputs = self.lightning_module.training_step_end(outputs)
        self.callback_handler.set_context_value(
            CallbackHandlerContextValues.TRAIN_OUTPUT,
            outputs if module_outputs is None else module_outputs,
        )
        outputs = self.callback_handler.on_train_step_end()
        return outputs

    @debug_log_on_call
    def backward(self, loss, optimizer, optimizer_idx):
        loss = self.callback_handler.on_backward_begin()
        self.lightning_module.backward(loss, optimizer, optimizer_idx)

    @debug_log_on_call
    def on_after_backward(self):
        self.lightning_module.on_after_backward()
        self.callback_handler.on_backward_end()

    @debug_log_on_call
    def validation_step(self, batch, batch_idx):
        self.callback_handler.set_context_value(
            CallbackHandlerContextValues.VALIDATION_BATCH, batch
        )
        batch = self.callback_handler.on_validation_batch_begin()
        loss = self.lightning_module.validation_step(batch, batch_idx)

        self.callback_handler.set_context_value(
            CallbackHandlerContextValues.VALIDATION_LOSS, loss.get("val_loss")
        )
        return loss

    @debug_log_on_call
    def validation_step_end(self, outputs):
        module_outputs = self.lightning_module.validation_step_end(outputs)
        self.callback_handler.set_context_value(
            CallbackHandlerContextValues.VALIDATION_OUTPUT,
            outputs if module_outputs is None else module_outputs,
        )
        outputs = self.callback_handler.on_validation_step_end()
        return outputs

    @debug_log_on_call
    def validation_epoch_end(self, outputs):
        module_outputs = self.lightning_module.validation_epoch_end(outputs)
        self.callback_handler.set_context_value(
            CallbackHandlerContextValues.ALL_VALIDATION_OUTPUT,
            outputs if module_outputs is None else module_outputs,
        )
        all_outputs = self.callback_handler.on_validation_epoch_end()
        return all_outputs

    @debug_log_on_call
    def training_epoch_end(self, outputs):
        module_outputs = self.lightning_module.training_epoch_end(outputs)
        self.callback_handler.set_context_value(
            CallbackHandlerContextValues.ALL_TRAIN_OUTPUT,
            outputs if module_outputs is None else module_outputs,
        )
        self.callback_handler.on_train_epoch_end()

    @debug_log_on_call
    def test_step(self, batch, batch_idx):
        self.callback_handler.set_context_value(
            CallbackHandlerContextValues.TEST_BATCH, batch
        )
        batch = self.callback_handler.on_test_batch_begin()
        loss = self.lightning_module.test_step(batch, batch_idx)
        self.callback_handler.set_context_value(
            CallbackHandlerContextValues.TEST_LOSS, loss.get("test_loss")
        )
        return loss

    @debug_log_on_call
    def test_step_end(self, outputs):
        module_outputs = self.lightning_module.test_step_end(outputs)
        self.callback_handler.set_context_value(
            CallbackHandlerContextValues.TEST_OUTPUT,
            outputs if module_outputs is None else module_outputs,
        )
        outputs = self.callback_handler.on_test_step_end()
        return outputs

    @debug_log_on_call
    def test_epoch_end(self, outputs):
        module_outputs = self.lightning_module.test_epoch_end(outputs)
        self.callback_handler.set_context_value(
            CallbackHandlerContextValues.ALL_TEST_OUTPUT,
            outputs if module_outputs is None else module_outputs,
        )
        all_outputs = self.callback_handler.on_test_epoch_end()
        return all_outputs


lookup = {
    "PytorchLightningModuleHooksToCallbackHandlerAdapter": PytorchLightningModuleHooksToCallbackHandlerAdapter
}



from torchvision.models.detection import fasterrcnn_resnet50_fpn, fcos_resnet50_fpn, retinanet_resnet50_fpn_v2

lookup = {
    "fasterrcnn_resnet50_fpn": fasterrcnn_resnet50_fpn,
    "fcos_resnet50_fpn": fcos_resnet50_fpn,
    "retinanet_resnet50_fpn_v2": retinanet_resnet50_fpn_v2,
}



def get_toolbox_lookup():
    from .core import lookup as core_lookup
    from .integrations import lookup as integrations_lookup
    from .pipeline import Pipeline
    from .pipeline.config_loader import load_config_from_path, load_config_from_paths, replace_config_variables
    from .pipeline.utils.lookup_functions import lookup as pipeline_lookup_functions
    from .utils import lookup as utils_lookup
    from .vision import lookup as vision_lookup

    lookup = {
        **core_lookup,
        **integrations_lookup,
        **vision_lookup,
        **pipeline_lookup_functions,
        **utils_lookup,
        "Pipeline": Pipeline,
        "load_config_from_path": load_config_from_path,
        "load_config_from_paths": load_config_from_paths,
        "replace_config_variables": replace_config_variables,
    }
    return lookup



from pathlib import Path

import cv2
import numpy as np

HAS_LIBVIPS = True
try:
    print("Importing pyvips")
    import pyvips
except Exception:
    HAS_LIBVIPS = False


def imread_rgb(path):
    if HAS_LIBVIPS:
        image = pyvips.Image.new_from_file(str(path), access="sequential")
        mem_img = image.write_to_memory()
        return np.frombuffer(mem_img, dtype=np.uint8).reshape(
            image.height, image.width, 3
        )
    else:
        return np.array(
            cv2.cvtColor(cv2.imread(str(path), cv2.IMREAD_UNCHANGED), cv2.COLOR_BGR2RGB)
        )


def imread_grayscale(path):
    if HAS_LIBVIPS:
        image = pyvips.Image.new_from_file(str(path), access="sequential")
        mem_img = image.write_to_memory()
        return np.frombuffer(mem_img, dtype=np.uint8).reshape(image.height, image.width)
    else:
        return np.array(cv2.imread(str(path), cv2.IMREAD_GRAYSCALE))


def imwrite(path, image):
    Path(path).parent.mkdir(exist_ok=True, parents=True)
    cv2.imwrite(path, image)


lookup = {
    "imread_rgb": imread_rgb,
    "imread_grayscale": imread_grayscale,
    "imwrite": imwrite,
}



from .image_bounding_box_detection import lookup as image_bounding_box_detection_lookup
from .image_classification import lookup as image_classification_lookup
from .image_segmentation import lookup as image_segmentation_lookup

lookup = {
    **image_bounding_box_detection_lookup,
    **image_segmentation_lookup,
    **image_classification_lookup,
}



from collections import defaultdict

import torch

from ....core.callbacks import Callback


class OnEpochEndTransformer(Callback):
    """
    Reduces the output from Pytorch Lightning Module from:

    [{"metric_A": "metric_A_sample_1", "metric_B": "metric_B_sample_1", ...}]

    to

    {
        metric_A: mean([metric_A_sample_1, metric_A_sample_2, ...])
        metric_B: mean([metric_B_sample_1, metric_B_sample_2, ...])
    }

    if self.reduce_keys in metric_A

    or:

    {
        metric_A: [metric_A_sample_1, metric_A_sample_2, ...]
    }

    if self.reduce_keys not in metric


    :param Callback: [description]
    :type Callback: [type]
    """

    def __init__(self, reduce_keys=["val_loss", "test_loss"], name=None):
        super().__init__(name)
        self.reduce_keys = reduce_keys

    def _reduce_output(self, all_output):
        reduced_output = defaultdict(list)
        for output in all_output:
            for k, v in output.items():
                reduced_output[k].append(v)
        return {
            k: torch.mean(torch.stack(v)) if k in self.reduce_keys else v
            for k, v in reduced_output.items()
        }

    def on_train_epoch_end(self, context):
        context.all_train_output = self._reduce_output(context.all_train_output)

    def on_validation_epoch_end(self, context):
        context.all_validation_output = self._reduce_output(
            context.all_validation_output
        )

    def on_test_epoch_end(self, context):
        context.all_test_output = self._reduce_output(context.all_test_output)


lookup = {"OnEpochEndTransformer": OnEpochEndTransformer}



import mlflow

from ...utils.conda import get_current_conda_env


def save_mlflow_pyfunc_model(
    model_save_path, model, artifacts=None, conda_env_name=None
):
    model_save_path = str(model_save_path)
    conda_env = get_current_conda_env(env_name=conda_env_name)
    mlflow.pyfunc.save_model(
        path=model_save_path,
        python_model=model,
        artifacts=artifacts,
        conda_env=conda_env,
    )


lookup = {"save_mlflow_pyfunc_model": save_mlflow_pyfunc_model}



# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/data_loader.ipynb (unless otherwise specified).

__all__ = [
    "DataLoader",
    "create_weighted_random_sampler",
    "create_sample_weights_from_multilabels",
    "lookup",
]

# Cell
from collections import Counter

from torch.utils.data import DataLoader as PytorchDataLoader
from torch.utils.data import Dataset, WeightedRandomSampler
from tqdm import tqdm


class DataLoader(PytorchDataLoader):
    pass


def create_weighted_random_sampler(weights, n_samples=None, replacement=True):
    if n_samples is None:
        n_samples = len(weights)
    return WeightedRandomSampler(
        weights, num_samples=n_samples, replacement=replacement
    )


def create_sample_weights_from_multilabels(multilabels, no_label_sample_factor=1):
    all_labels = []
    labels_for_samples = []
    sample_weights = []
    n_samples = len(multilabels)

    for labels in tqdm(multilabels):
        labels_for_samples.append(labels)
        all_labels.extend(labels)

    label_counts = Counter(all_labels)

    weight_lookup = {label: 1 / n_labels for label, n_labels in label_counts.items()}

    for labels_for_sample in labels_for_samples:
        if len(labels_for_sample) == 0:
            sample_weights.append(1 / n_samples ** no_label_sample_factor)
            continue
        sample_weights.append(
            sum([weight_lookup[label] for label in labels_for_sample])
            / len(labels_for_sample)
        )
    return sample_weights


lookup = {
    "DataLoader": DataLoader,
    "create_weighted_random_sampler": create_weighted_random_sampler,
    "create_sample_weights_from_multilabels": create_sample_weights_from_multilabels,
}



from .label_statistics import *
from .label_statistics import lookup as label_statistics_lookup
from .mean_average_precision import *
from .mean_average_precision import lookup as mean_average_precision_lookup

lookup = {**mean_average_precision_lookup, **label_statistics_lookup}



import mlflow


def load_mlflow_pyfunc_model(model_save_path):
    from ....engine import get_toolbox_lookup

    model_save_path = str(model_save_path)
    model = mlflow.pyfunc.load_model(model_save_path)
    python_model = model._model_impl.python_model
    python_model.lookup = get_toolbox_lookup()
    return model


lookup = {"load_mlflow_pyfunc_model": load_mlflow_pyfunc_model}



from ......vision.utils import imread_rgb
from .sample import ImageBoundingBoxDetectionSample


def create_image_bboxes_detection_sample(
    image_data=None,
    image_path=None,
    load_image_fn=imread_rgb,
    labels_data=None,
    labels_path=None,
    load_labels_fn=None,
    bboxes_data=None,
    bboxes_path=None,
    load_bboxes_fn=None,
    bboxes_format=None,
    transform_to_bboxes_format=None,
    labels_to_keep=None,
    transform_label_map=None,
    name=None,
):
    """Filtering is done first, hence the label map should be using the original labels

    Args:
        image_data (_type_, optional): _description_. Defaults to None.
        image_path (_type_, optional): _description_. Defaults to None.
        load_image_fn (_type_, optional): _description_. Defaults to imread_rgb.
        labels_data (_type_, optional): _description_. Defaults to None.
        labels_path (_type_, optional): _description_. Defaults to None.
        load_labels_fn (_type_, optional): _description_. Defaults to None.
        bboxes_data (_type_, optional): _description_. Defaults to None.
        bboxes_path (_type_, optional): _description_. Defaults to None.
        load_bboxes_fn (_type_, optional): _description_. Defaults to None.
        bboxes_format (_type_, optional): _description_. Defaults to None.
        transform_to_bboxes_format (_type_, optional): _description_. Defaults to None.
        labels_to_keep (_type_, optional): _description_. Defaults to None.
        transform_label_map (_type_, optional): _description_. Defaults to None.
        name (_type_, optional): _description_. Defaults to None.

    Returns:
        _type_: _description_
    """
    sample = ImageBoundingBoxDetectionSample.create(
        image_data,
        image_path,
        load_image_fn,
        labels_data,
        labels_path,
        load_labels_fn,
        bboxes_data,
        bboxes_path,
        load_bboxes_fn,
        bboxes_format,
        name,
    )

    if transform_to_bboxes_format == "PASCAL_VOC":
        sample.convert_to_pascal_voc_bboxes(lazy=True)

    if labels_to_keep is not None:
        sample.filter_labels_and_bboxes_to_keep(labels_to_keep, lazy=True)

    if transform_label_map is not None:
        sample.transform_labels(transform_label_map, lazy=True)

    return sample


lookup = {"create_image_bboxes_detection_sample": create_image_bboxes_detection_sample}



import pretrainedmodels

from .change_last_linear import cadene_change_last_linear

lookup = {
    "fbresnet152": pretrainedmodels.models.fbresnet.fbresnet152,
    "bninception": pretrainedmodels.models.bninception,
    "resnext101_32x4d": pretrainedmodels.models.resnext.resnext101_32x4d,
    "resnext101_64x4d": pretrainedmodels.models.resnext.resnext101_64x4d,
    "inceptionv4": pretrainedmodels.models.inceptionv4,
    "inceptionresnetv2": pretrainedmodels.models.inceptionresnetv2,
    "alexnet": pretrainedmodels.models.torchvision_models.alexnet,
    "densenet121": pretrainedmodels.models.torchvision_models.densenet121,
    "densenet169": pretrainedmodels.models.torchvision_models.densenet169,
    "densenet201": pretrainedmodels.models.torchvision_models.densenet201,
    "densenet161": pretrainedmodels.models.torchvision_models.densenet161,
    "resnet18": pretrainedmodels.models.torchvision_models.resnet18,
    "resnet34": pretrainedmodels.models.torchvision_models.resnet34,
    "resnet50": pretrainedmodels.models.torchvision_models.resnet50,
    "resnet101": pretrainedmodels.models.torchvision_models.resnet101,
    "resnet152": pretrainedmodels.models.torchvision_models.resnet152,
    "inceptionv3": pretrainedmodels.models.torchvision_models.inceptionv3,
    "squeezenet1_0": pretrainedmodels.models.torchvision_models.squeezenet1_0,
    "squeezenet1_1": pretrainedmodels.models.torchvision_models.squeezenet1_1,
    "vgg11": pretrainedmodels.models.torchvision_models.vgg11,
    "vgg11_bn": pretrainedmodels.models.torchvision_models.vgg11_bn,
    "vgg13": pretrainedmodels.models.torchvision_models.vgg13,
    "vgg13_bn": pretrainedmodels.models.torchvision_models.vgg13_bn,
    "vgg16": pretrainedmodels.models.torchvision_models.vgg16,
    "vgg16_bn": pretrainedmodels.models.torchvision_models.vgg16_bn,
    "vgg19_bn": pretrainedmodels.models.torchvision_models.vgg19_bn,
    "vgg19": pretrainedmodels.models.torchvision_models.vgg19,
    "nasnetamobile": pretrainedmodels.models.nasnet_mobile.nasnetamobile,
    "nasnetalarge": pretrainedmodels.models.nasnet.nasnetalarge,
    "dpn68": pretrainedmodels.models.dpn.dpn68,
    "dpn68b": pretrainedmodels.models.dpn.dpn68b,
    "dpn92": pretrainedmodels.models.dpn.dpn92,
    "dpn98": pretrainedmodels.models.dpn.dpn98,
    "dpn131": pretrainedmodels.models.dpn.dpn131,
    "dpn107": pretrainedmodels.models.dpn.dpn107,
    "xception": pretrainedmodels.models.xception,
    "senet154": pretrainedmodels.models.senet.senet154,
    "se_resnet50": pretrainedmodels.models.senet.se_resnet50,
    "se_resnet101": pretrainedmodels.models.senet.se_resnet101,
    "se_resnet152": pretrainedmodels.models.senet.se_resnet152,
    "se_resnext50_32x4d": pretrainedmodels.models.senet.se_resnext50_32x4d,
    "se_resnext101_32x4d": pretrainedmodels.models.senet.se_resnext101_32x4d,
    "cafferesnet101": pretrainedmodels.models.cafferesnet.cafferesnet101,
    "pnasnet5large": pretrainedmodels.models.pnasnet.pnasnet5large,
    "polynet": pretrainedmodels.models.polynet,
    "cadene_change_last_linear": cadene_change_last_linear,
}



from .augmentations import *
from .augmentations import augmentation_lookup
from .deployment import *
from .deployment import lookup as deployment_lookup
from .models import *
from .models import models_lookup
from .tasks import *
from .tasks import lookup as tasks_lookup
from .utils import *
from .utils import lookup as utils_lookup

lookup = {
    **augmentation_lookup,
    **models_lookup,
    **deployment_lookup,
    **tasks_lookup,
    **utils_lookup,
}



import logging
import sys
from contextlib import ContextDecorator

LOGGER_NAME = "engine"


def create_logger():
    logger = logging.getLogger(LOGGER_NAME)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setLevel(logging.DEBUG)
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)
    return logger


logger = create_logger()


def set_log_level(level=logging.INFO):
    logging.getLogger(LOGGER_NAME).setLevel(level)


def set_no_logs():
    logging.getLogger(LOGGER_NAME).setLevel(sys.maxsize)


def set_info_logs():
    logging.getLogger(LOGGER_NAME).setLevel(logging.INFO)


def set_debug_logs():
    logging.getLogger(LOGGER_NAME).setLevel(logging.DEBUG)


class debug_engine_logs(ContextDecorator):
    def __enter__(self):
        self.previous_log_level = logger.level
        set_debug_logs()
        return self

    def __exit__(self, *exc):
        set_log_level(self.previous_log_level)
        return False


class info_engine_logs(ContextDecorator):
    def __enter__(self):
        self.previous_log_level = logger.level
        set_info_logs()
        return self

    def __exit__(self, *exc):
        set_log_level(self.previous_log_level)
        return False


class no_engine_logs(ContextDecorator):
    def __enter__(self):
        self.previous_log_level = logger.level
        set_no_logs()
        return self

    def __exit__(self, *exc):
        set_log_level(self.previous_log_level)
        return False



import numpy as np


class LabelStatisticsForMeanAveragePrecision:
    def __init__(self, label):
        self.label = label
        self.false_positives = []
        self.true_positives = []
        self.scores = []
        self.num_annotations = 0

    @property
    def true_positives_sorted_by_confidence_cumulative_sum(self):
        true_positives_sorted_by_confidence = np.array(self.true_positives)[
            np.argsort(-np.array(self.scores))
        ]
        true_positives_sorted_by_confidence_cumulative_sum = np.cumsum(
            true_positives_sorted_by_confidence
        )
        return true_positives_sorted_by_confidence_cumulative_sum

    @property
    def false_positives_sorted_by_confidence_cumulative_sum(self):
        false_positives_sorted_by_confidence = np.array(self.false_positives)[
            np.argsort(-np.array(self.scores))
        ]
        false_positives_sorted_by_confidence_cumulative_sum = np.cumsum(
            false_positives_sorted_by_confidence
        )
        return false_positives_sorted_by_confidence_cumulative_sum

    @property
    def recall(self):
        return (
            self.true_positives_sorted_by_confidence_cumulative_sum
            / self.num_annotations
        )

    @property
    def precision(self):
        return self.true_positives_sorted_by_confidence_cumulative_sum / np.maximum(
            self.true_positives_sorted_by_confidence_cumulative_sum
            + self.false_positives_sorted_by_confidence_cumulative_sum,
            np.finfo(np.float64).eps,
        )

    @property
    def average_precision(self):
        return self._compute_ap(self.recall, self.precision)

    def _compute_ap(self, recall, precision):
        """Compute the average precision, given the recall and precision curves.
        Code originally from https://github.com/rbgirshick/py-faster-rcnn.
        # Arguments
            recall:    The recall curve (list).
            precision: The precision curve (list).
        # Returns
            The average precision as computed in py-faster-rcnn.
        """
        # correct AP calculation
        # first append sentinel values at the end
        mrec = np.concatenate(([0.0], recall, [1.0]))
        mpre = np.concatenate(([0.0], precision, [0.0]))

        # compute the precision envelope
        for i in range(mpre.size - 1, 0, -1):
            mpre[i - 1] = np.maximum(mpre[i - 1], mpre[i])

        # to calculate area under PR curve, look for points
        # where X axis (recall) changes value
        i = np.where(mrec[1:] != mrec[:-1])[0]

        # and sum (\Delta recall) * prec
        ap = np.sum((mrec[i + 1] - mrec[i]) * mpre[i + 1])
        return ap


lookup = {
    "LabelStatisticsForMeanAveragePrecision": LabelStatisticsForMeanAveragePrecision
}



from .bbox import *
from .bbox import lookup as bbox_lookup
from .image import *
from .image import lookup as image_lookup
from .normalization import *
from .normalization import lookup as normalization_lookup
from .rle import *
from .rle import lookup as rle_lookup
from .transforms import *
from .transforms import lookup as transform_lookup
from .u_net_weight_map import *

lookup = {
    **bbox_lookup,
    **image_lookup,
    **normalization_lookup,
    **rle_lookup,
    **transform_lookup,
}



import subprocess
from contextlib import contextmanager
from tempfile import NamedTemporaryFile
from urllib.parse import unquote

import boto3
from tqdm import tqdm

# https://stackoverflow.com/questions/43458001/can-we-copy-the-files-and-folders-recursively-between-aws-s3-buckets-using-boto3/43478147


class S3Client:
    def __init__(self, profile_name=None, endpoint_url=None):
        session_kwargs = {}
        client_kwargs = {}
        if profile_name is not None:
            session_kwargs["profile_name"] = profile_name
        if endpoint_url is not None:
            client_kwargs["endpoint_url"] = endpoint_url
        self.client = boto3.session.Session(**session_kwargs).client(
            "s3", **client_kwargs
        )

    @contextmanager
    def download_to_temporary_file(self, bucket, key, show_progress=False):
        try:
            temp_file = NamedTemporaryFile()
            self.download(bucket, key, temp_file.name, show_progress)
            yield temp_file.name
        finally:
            pass

    def download(self, bucket, key, save_path, show_progress=False):
        save_path = str(save_path)

        if show_progress:

            def hook(t):
                def inner(bytes_amount):
                    t.update(bytes_amount)

                return inner

            file_size = float(
                self.client.head_object(Bucket=bucket, Key=unquote(key))[
                    "ContentLength"
                ]
            )
            with tqdm(total=file_size, unit="B", unit_scale=True, desc=save_path) as t:
                self.client.download_file(
                    bucket, unquote(key), save_path, Callback=hook(t)
                )
        else:
            self.client.download_file(bucket, unquote(key), save_path)

    def upload_file(self, path, bucket, key):
        self.client.upload_file(path, bucket, key)

    def upload_fileobj(self, fileobj, bucket, key):
        self.client.upload_fileobj(fileobj, bucket, key)

    def copy(self, source_bucket, source_key, dst_bucket, dst_key):
        self.client.copy(
            {"Bucket": source_bucket, "Key": source_key}, dst_bucket, dst_key
        )

    def sync(self, from_path, to_path):
        cmd = ["aws"]

        if self.endpoint_url is not None:
            cmd.append(f"--endpoint-url {self.endpoint_url}")

        if self.profile_name is not None:
            cmd.append(f"=--profile {self.profile_name}")

        cmd.append(from_path)
        cmd.append(to_path)
        return subprocess.Popen(cmd).wait()

    def check_if_exists_in_s3(self, bucket, key):
        try:
            self.client.head_object(Bucket=bucket, Key=key)
            return True
        except Exception:
            return False

    def get_s3_keys(self, bucket, prefix=None, delimiter=None):
        keys = []
        try:
            for resp in self._construct_s3_paginator(bucket, prefix, delimiter):
                keys.extend([c["Key"] for c in resp["Contents"]])
        except KeyError:
            pass
        return keys

    def get_top_level_bucket_keys(self, bucket):
        keys = []
        for prefix in self._construct_s3_paginator(bucket, delimiter="/").search(
            "CommonPrefixes"
        ):
            keys.append(prefix.get("Prefix"))
        return keys

    def _construct_s3_paginator(self, bucket, prefix=None, delimiter=None):
        kwargs = {"Bucket": bucket}
        s3_paginator = self.client.get_paginator("list_objects_v2")

        if prefix is not None:
            kwargs.update({"Prefix": prefix})

        if delimiter is not None:
            kwargs.update({"Delimiter": delimiter})
        return s3_paginator.paginate(**kwargs)



import cv2
import numpy as np
from albumentations.augmentations import Crop, PadIfNeeded

from ....utils.convenience import find_nearest_greater_divisible


def create_tiled_square_crops(array, size=256):
    padded_array = create_padded_array(array, size)
    crop_areas = create_square_crop_areas(padded_array, size=size)
    cropped_arrays = create_crops(crop_areas, padded_array)
    return cropped_arrays


def merge_tiled_square_crops(tiles, original_height, original_width, size=256):
    padded_height = find_nearest_greater_divisible(original_height, size)
    padded_width = find_nearest_greater_divisible(original_width, size)
    n_tiles_in_height = padded_height // size
    n_tiles_in_width = padded_width // size
    n_tiles_per_image = n_tiles_in_height * n_tiles_in_width
    merged_images = []
    for i in range(0, len(tiles), n_tiles_per_image):
        merged_images.append(
            merge_tiled_square_crops_for_single(
                tiles[i : i + n_tiles_per_image], padded_height, padded_width, size
            )
        )
    return merged_images


def merge_tiled_square_crops_for_single(
    tiles_for_image, original_height, original_width, size=256
):
    merged_horizontal_tiles = []
    n_tiles_in_height = original_height // size
    n_tiles_in_width = original_width // size
    for i in range(0, len(tiles_for_image), n_tiles_in_width):
        merged_horizontal_tiles.append(
            np.hstack(tiles_for_image[i : i + n_tiles_in_width])
        )
    assert len(merged_horizontal_tiles) == n_tiles_in_height
    return np.vstack(merged_horizontal_tiles)


def create_padded_array(array, size=256):
    height, width, *_ = array.shape
    padded = PadIfNeeded(
        find_nearest_greater_divisible(height, size),
        find_nearest_greater_divisible(width, size),
        border_mode=cv2.BORDER_CONSTANT,
        value=0,
    )(image=array)["image"]
    return padded


def create_crops(crop_areas, array):
    cropped_arrays = []
    for crop_area in crop_areas:
        cropped = Crop(**crop_area)(image=array)
        cropped_arrays.append(cropped["image"])
    return cropped_arrays


def create_square_crop_areas(array, size):
    crops = []
    height, width, *_ = array.shape
    for i in range(height // size):
        for j in range(width // size):
            x_min = j * size
            y_min = i * size
            x_max = (j + 1) * size
            y_max = (i + 1) * size
            crops.append(
                {"x_min": x_min, "y_min": y_min, "x_max": x_max, "y_max": y_max}
            )
            assert x_max - x_min == size
            assert y_max - y_min == size
    return crops


lookup = {
    "create_tiled_square_crops": create_tiled_square_crops,
    "merge_tiled_square_crops": merge_tiled_square_crops,
    "create_square_crop_areas": create_square_crop_areas,
    "create_crops": create_crops,
    "create_padded_array": create_padded_array,
    "merge_tiled_square_crops_for_single": merge_tiled_square_crops_for_single,
}



from ...utils.convenience import run_shell_command


def get_istio_ingress_ip():
    return run_shell_command(
        "microk8s kubectl get po -l istio=ingressgateway -n istio-system -o jsonpath='{.items[0].status.hostIP}'",
        silent=True,
    )[0]


def get_istio_ingress_port():
    return run_shell_command(
        "microk8s kubectl -n istio-system get service istio-ingressgateway -o jsonpath='{.spec.ports[?(@.name==\"http2\")].nodePort'}",
        silent=True,
    )[0]



from .create import *
from .create import lookup as create_lookup
from .filter import *
from .filter import lookup as filter_lookup
from .sample import *
from .sample import lookup as sample_lookup

lookup = {**create_lookup, **filter_lookup, **sample_lookup}



import numpy as np


# ref.: https://www.kaggle.com/stainsby/fast-tested-rle
def rle_encode(mask):
    pixels = mask.T.flatten()
    # We need to allow for cases where there is a '1' at either end of the sequence.
    # We do this by padding with a zero at each end when needed.
    use_padding = False
    if pixels[0] or pixels[-1]:
        use_padding = True
        pixel_padded = np.zeros([len(pixels) + 2], dtype=pixels.dtype)
        pixel_padded[1:-1] = pixels
        pixels = pixel_padded
    rle = np.where(pixels[1:] != pixels[:-1])[0] + 2
    if use_padding:
        rle = rle - 1
    rle[1::2] = rle[1::2] - rle[:-1:2]
    return " ".join(map(str, rle.tolist()))


# This is copied from https://www.kaggle.com/paulorzp/run-length-encode-and-decode.
def rle_decode(rle_str, mask_shape, mask_dtype=np.uint8):
    s = rle_str.split()
    starts, lengths = [np.asarray(x, dtype=int) for x in (s[0:][::2], s[1:][::2])]
    starts -= 1
    ends = starts + lengths
    mask = np.zeros(np.prod(mask_shape), dtype=mask_dtype)
    for lo, hi in zip(starts, ends):
        mask[lo:hi] = 1
    return mask.reshape(mask_shape[::-1]).T


lookup = {"rle_encode": rle_encode, "rle_decode": rle_decode}
"""
Test

img = np.array([[0,0,0,1,0],
                [0,0,1,1,0],
                [1,0,0,0,1],
                [1,1,0,0,1]])

rle = rle_encode(img)

decode_img = rle_decode(rle, shape=img.shape)

assert img == decode_img

"""



import mlflow
from pytorch_lightning.utilities.cloud_io import load as pl_load

from ....engine import get_toolbox_lookup


class MLFlowImageModel(mlflow.pyfunc.PythonModel):
    def load_context(self, context):
        self.lookup = get_toolbox_lookup()
        self.config = self.lookup["load_config_from_path"](
            context.artifacts["config_path"]
        )
        self.config["Variables"]["is_deployed"] = True
        self.pipeline = self.lookup["Pipeline"].create_from_config(
            self.config, self.lookup
        )
        self.pipeline.run(remove_nodes_with_tags=["TRAIN"])
        self.module = self.pipeline.get_node_output("ModuleCreator")()
        self.trainer = self.pipeline.get_node_output("TrainerCreator")()
        self.inference_recorder = self.pipeline.get_node_output("InferenceRecorder")

        pl_checkpoint = pl_load(
            context.artifacts["model_save_path"],
            map_location=lambda storage, loc: storage,
        )
        self.module.load_state_dict(pl_checkpoint["state_dict"])

    def predict(self, context, model_input):
        if "images" in model_input:
            samples_data = [
                {"image_data": self.lookup["to_uint8_image"](image)}
                for image in model_input["images"]
            ]
        if "image_paths" in model_input:
            samples_data = [
                {
                    "image_data": self.lookup["to_uint8_image"](
                        self.lookup["imread_rgb"](image_path)
                    )
                }
                for image_path in model_input["image_paths"]
            ]
        samples = [
            self.pipeline.get_node_output("SampleFromDataCreator")(**sample_data)
            for sample_data in samples_data
        ]
        test_dataset = self.pipeline.get_node_output("TestDatasetCreator")(
            samples=samples
        )
        test_dataloader = self.pipeline.get_node_output("TestDataLoaderCreator")(
            test_dataset
        )
        self.trainer.test(self.module, dataloaders=[test_dataloader])
        predictions = self.inference_recorder.results
        self.inference_recorder.flush_results()
        return predictions


lookup = {"MLFlowImageModel": MLFlowImageModel}



import importlib.util
import inspect
import pprint
from pathlib import Path

from ...logger import logger


def create_lookup(path, package_name, ignore_paths=[]):
    """Start at `path` and traverses the directory tree and creates a key, value pair lookup for all the
    classes and functions found in *.py files, except for files in ignore_paths

    :param path: root path to start the traversal
    :type path: str or Path
    :param package_name: name of the package for the lookup to be generated, this is needed to generate lookups for relative importss
    :type package_name: str
    :param ignore_paths: paths to ignore for lookup creation, defaults to []
    :type ignore_paths: list, optional
    :return: a lookup dictionary that can be used by the Pipeline
    :rtype: dict
    """
    lookup = {}
    for p in Path(path).glob("*"):
        if p in ignore_paths:
            lookup = {**lookup}
            logger.info(f"Ignoring file: {str(p)}")
        elif p.suffix == ".py":
            logger.info(f"Found file: {str(p)} and creating lookup")
            created_lookup = create_lookup_from_path(p, package_name)
            logger.info(f"Created lookup: {pprint.pformat(created_lookup)}")
            lookup = {**lookup, **created_lookup}
        elif p.is_dir():
            logger.info(f"Found folder: {str(p)}")
            lookup = {**lookup, **create_lookup(p, package_name, ignore_paths)}
        else:
            logger.info(f"Nothing here: {str(p)}")
            lookup = {**lookup}
    # logger.info(f"Current lookup is: {pprint.pformat(lookup)}")

    return lookup


def create_lookup_from_path(path, package_name):
    try:
        # logger.info(f"Creating lookup from path: {str(path)}")
        module = load_module_from_path(path, package_name)
        lookup = create_lookup_from_module(module)
        # logger.info(f"Lookup generated: {lookup}")
        return lookup
    except ImportError as e:
        if str(e) == "attempted relative import with no known parent package":
            logger.info(f"Importing of module from path: {path} has failed")
            logger.info(e)
            return {}
        else:
            logger.error(f"Importing of module from path: {path} has failed")
            logger.error(e)
            raise ImportError


def load_module_from_path(path, package_name):
    """Create a module the path

    For example:

    path = Path(/home/kevin/Documents/Projects/pytorch-toolbox/pytorch_toolbox/pipeline/utils/saving.py)
    module_name = saving
    full_module_name = pytorch_toolbox.pipeline.utils.saving

    :param path: Full path to a module
    :type path: Path
    :param package_name: name of the package for the lookup to be generated
    :type package_name: str

    :return: Module
    :rtype: Module
    :return: [description]
    :rtype: [type]
    """

    # Take the last index to avoid edge case where there is another folder named pytorch_toolbox in the whole path
    package_index = [
        i for i, p in enumerate(path.absolute().parts) if p == package_name
    ][-1]
    module_name = path.stem
    full_module_name = (
        f"{'.'.join(list(path.absolute().parts[package_index:-1]))}.{module_name}"
    )
    module = create_module_from_path(path, module_name=full_module_name)
    return module


def create_module_from_path(path, module_name):
    spec = importlib.util.spec_from_file_location(module_name, str(path))
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def create_lookup_from_module(module):
    function_names, function_pointers = load_functions_from_modules(module)
    classes_names, classes_pointers = load_classes_from_modules(module)
    return {
        name: pointer
        for (name, pointer) in zip(
            function_names + classes_names, function_pointers + classes_pointers
        )
    }


def load_functions_from_modules(module):
    try:
        function_names, function_pointers = zip(
            *[
                member
                for member in inspect.getmembers(module, inspect.isfunction)
                if member[1].__module__ == module.__name__
            ]
        )
        return function_names, function_pointers
    except ValueError:  # if there are no functions defined
        return (), ()


def load_classes_from_modules(module):
    try:
        class_names, class_pointers = zip(
            *[
                member
                for member in inspect.getmembers(module, inspect.isclass)
                if member[1].__module__ == module.__name__
            ]
        )
        return class_names, class_pointers
    except ValueError:  # if there are no classes defined
        return (), ()


def generate_imports_from_lookup(lookup):
    lookup_with_import_path = {}
    modules_to_be_imported = []
    for name, pointer in lookup.items():
        pointer_module = pointer.__module__
        module_name = pointer_module.split(".")[0]
        lookup_with_import_path[name] = f"{pointer_module}.{name}"
        if module_name not in modules_to_be_imported:
            modules_to_be_imported.append(module_name)
    return lookup_with_import_path, modules_to_be_imported



import subprocess

import yaml


def get_current_conda_env(env_name=None):
    command = ["conda", "env", "export", "--no-builds"]
    if env_name:
        command.extend(["--name", env_name])
    process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = process.communicate()
    return yaml.load(stdout.decode("UTF-8"), Loader=yaml.Loader)


lookup = {"get_current_conda_env": get_current_conda_env}



import pytorch_lightning as pl

from ...integrations.pytorch_lightning.adapters import PytorchLightningCallbacksToCallbackHandlerAdapter


def create_pytorch_lightning_trainer(
    callback_handler,
    gpus=1,
    max_epochs=1,
    checkpoint_callback=None,
    resume_from_checkpoint=None,
    gradient_clip_val=0,
    num_sanity_val_steps=0,
    precision=16,
):

    return pl.Trainer(
        callbacks=[PytorchLightningCallbacksToCallbackHandlerAdapter(callback_handler)],
        gpus=gpus,
        max_epochs=max_epochs,
        resume_from_checkpoint=resume_from_checkpoint,
        checkpoint_callback=checkpoint_callback,
        gradient_clip_val=gradient_clip_val,
        num_sanity_val_steps=num_sanity_val_steps,
        precision=precision,
        logger=None,
    )


lookup = {"create_pytorch_lightning_trainer": create_pytorch_lightning_trainer}



from .core import *

lookup = {
    "unet_with_resnet18_backbone": unet_with_resnet18_backbone,
    "unet_with_resnet34_backbone": unet_with_resnet34_backbone,
    "unet_with_resnet50_backbone": unet_with_resnet50_backbone,
    "unet_with_resnet101_backbone": unet_with_resnet101_backbone,
    "unet_with_resnet152_backbone": unet_with_resnet152_backbone,
}



import albumentations
import cv2
import numpy as np
from albumentations.augmentations import Crop, PadIfNeeded

from ....utils.convenience import find_nearest_greater_divisible
from ....vision.core.augmentations import create_augmentation_fn


def create_tiled_square_crops_for_image_with_bboxes(image, bboxes, labels, size=256):
    image, bboxes, labels = create_padded_image_with_bboxes(image, bboxes, labels, size)
    tile_coordinates = create_square_tile_coordinates(image, size=size)
    tiled_images_with_bboxes = create_tiles_for_image_with_bboxes(
        tile_coordinates, image, bboxes, labels
    )
    return tiled_images_with_bboxes


def create_padded_image_with_bboxes(image, bboxes, labels, size=256):
    height, width, *_ = image.shape
    output = create_augmentation_fn(
        aug_fns=[
            PadIfNeeded(
                find_nearest_greater_divisible(height, size),
                find_nearest_greater_divisible(width, size),
                border_mode=cv2.BORDER_CONSTANT,
                value=0,
            )
        ],
        bbox_params=albumentations.BboxParams(
            format="pascal_voc", label_fields=["labels"]
        ),
    )(image=image, bboxes=bboxes, labels=labels)
    return output["image"], output["bboxes"], output["labels"]


def create_square_tile_coordinates(array, size):
    crops = []
    height, width, *_ = array.shape
    for i in range(height // size):
        for j in range(width // size):
            x_min = j * size
            y_min = i * size
            x_max = (j + 1) * size
            y_max = (i + 1) * size
            crops.append(
                {"x_min": x_min, "y_min": y_min, "x_max": x_max, "y_max": y_max}
            )
            assert x_max - x_min == size
            assert y_max - y_min == size
    return crops


def create_tiles_for_image_with_bboxes(tile_coordinates, image, bboxes, labels):
    tiled_image_with_bboxes = []
    for tile_coordinate in tile_coordinates:
        output = create_augmentation_fn(
            aug_fns=[Crop(**tile_coordinate)],
            bbox_params=albumentations.BboxParams(
                format="pascal_voc", label_fields=["labels"]
            ),
        )(image=image, bboxes=bboxes, labels=labels)
        tiled_image_with_bboxes.append(
            (output["image"], output["bboxes"], output["labels"])
        )
    return tiled_image_with_bboxes



from .box_iou import *
from .box_iou import lookup as box_iou_lookup
from .core import *
from .core import lookup as core_lookup
from .mean_average_precision import *
from .mean_average_precision import lookup as mean_average_precision_lookup

lookup = {**core_lookup, **box_iou_lookup, **mean_average_precision_lookup}



import torch
import torch.utils

from .....vision.tasks.image_segmentation.data.sample import create_image_segmentation_sample


class SegmentationAugmentationMixin:
    def augment(self, image, mask=None, weight_map=None):
        if self.augment_fn is None:
            return image, mask, weight_map

        if mask is None and weight_map is None:
            return self.augment_fn(image=image)["image"], None, None

        if mask is not None and weight_map is not None:
            augmented = self.augment_fn(image=image, masks=[mask, weight_map])
            return augmented["image"], augmented["masks"][0], augmented["masks"][1]

        if mask is not None:
            augmented = self.augment_fn(image=image, masks=[mask])
            return augmented["image"], augmented["masks"][0], None

        raise NotImplementedError(
            "Nothing has been implemented for a combination of images and weight_maps"
        )


class ImageSegmentationDatasetV2(
    torch.utils.data.Dataset, SegmentationAugmentationMixin
):
    def __init__(
        self,
        samples,
        augment_fn=None,
        tile_fn=None,
        convert_image_to_tensor_fn=lambda image: torch.from_numpy(
            image.transpose(2, 0, 1)
        ).float(),
        convert_mask_to_tensor_fn=lambda mask: torch.from_numpy(
            mask.transpose(2, 0, 1)
        ).float(),
    ):
        super().__init__()
        self.samples = samples
        self.augment_fn = augment_fn
        self.tile_fn = tile_fn
        self.convert_image_to_tensor_fn = convert_image_to_tensor_fn
        self.convert_mask_to_tensor_fn = convert_mask_to_tensor_fn

    def __len__(self):
        return len(self.samples)

    def __getitem__(self, i):
        sample = self.samples[i]
        image, mask, weight_map, _ = sample.data
        image, mask, weight_map = self.augment(image, mask, weight_map)

        sample.image.data = image

        if self.tile_fn is not None:
            image_tiles = self.tile_fn(image)

        if self.convert_image_to_tensor_fn is not None:
            if self.tile_fn is not None:
                sample.image.tensor = torch.stack(
                    [
                        self.convert_image_to_tensor_fn(image_tile)
                        for image_tile in image_tiles
                    ]
                )
            else:
                sample.image.tensor = self.convert_image_to_tensor_fn(image)

        if sample.has_mask:

            sample.mask.data = mask

            if self.tile_fn is not None:
                mask_tiles = self.tile_fn(mask)

            if self.convert_mask_to_tensor_fn is not None:
                if self.tile_fn is not None:
                    sample.mask.tensor = torch.stack(
                        [
                            self.convert_mask_to_tensor_fn(mask_tile)
                            for mask_tile in mask_tiles
                        ]
                    )
                else:
                    sample.mask.tensor = self.convert_mask_to_tensor_fn(mask)
        return sample


class ImageSegmentationDataset(torch.utils.data.Dataset, SegmentationAugmentationMixin):
    def __init__(
        self,
        samples_data,
        create_sample_fn=lambda sample_data: create_image_segmentation_sample(
            **sample_data
        ),
        augment_fn=None,
        tile_fn=None,
        convert_image_to_tensor_fn=lambda image: torch.from_numpy(
            image.transpose(2, 0, 1)
        ).float(),
        convert_mask_to_tensor_fn=lambda mask: torch.from_numpy(
            mask.transpose(2, 0, 1)
        ).float(),
    ):
        super().__init__()
        self.samples_data = samples_data
        self.create_sample_fn = create_sample_fn
        self.augment_fn = augment_fn
        self.tile_fn = tile_fn
        self.convert_image_to_tensor_fn = convert_image_to_tensor_fn
        self.convert_mask_to_tensor_fn = convert_mask_to_tensor_fn

    def __len__(self):
        return len(self.samples_data)

    def __getitem__(self, i):
        sample = self.create_sample_fn(self.samples_data[i])
        image, mask, weight_map, _ = sample.data
        image, mask, weight_map = self.augment(image, mask, weight_map)

        sample.image.data = image

        if self.tile_fn is not None:
            image_tiles = self.tile_fn(image)

        if self.convert_image_to_tensor_fn is not None:
            if self.tile_fn is not None:
                sample.image.tensor = torch.stack(
                    [
                        self.convert_image_to_tensor_fn(image_tile)
                        for image_tile in image_tiles
                    ]
                )
            else:
                sample.image.tensor = self.convert_image_to_tensor_fn(image)

        if sample.has_mask:

            sample.mask.data = mask

            if self.tile_fn is not None:
                mask_tiles = self.tile_fn(mask)

            if self.convert_mask_to_tensor_fn is not None:
                if self.tile_fn is not None:
                    sample.mask.tensor = torch.stack(
                        [
                            self.convert_mask_to_tensor_fn(mask_tile)
                            for mask_tile in mask_tiles
                        ]
                    )
                else:
                    sample.mask.tensor = self.convert_mask_to_tensor_fn(mask)
        return sample


def image_segmentation_collate_fn(samples):
    """Hook to be called before returning the data to the DataLoader

    :param sample: [description]
    :type sample: dict
    """

    batch = {}
    has_tiles = len(samples[0].image.tensor.shape) == 4
    has_mask = samples[0].has_mask

    batch["names"] = {
        "data": torch.utils.data.dataloader.default_collate([s.name for s in samples])
    }

    batch["images"] = (
        torch.utils.data.dataloader.default_collate([s.image.tensor for s in samples])
        if not has_tiles
        else torch.vstack([s.image.tensor for s in samples])
    )

    if has_mask:
        batch["masks"] = (
            torch.utils.data.dataloader.default_collate(
                [s.mask.tensor for s in samples]
            )
            if not has_tiles
            else torch.vstack([s.mask.tensor for s in samples])
        )
    return batch


lookup = {
    "ImageSegmentationDataset": ImageSegmentationDataset,
    "ImageSegmentationDatasetV2": ImageSegmentationDatasetV2,
    "image_segmentation_collate_fn": image_segmentation_collate_fn,
}



from ..callback import Callback


class CallbackMethodIsCalledRecorder(Callback):
    def __init__(self, name=None):
        super().__init__(name)
        self.method_is_called = {
            "on_setup": True,
            "on_init_begin": False,
            "on_init_end": False,
            "on_train_begin": False,
            "on_train_epoch_begin": False,
            "on_train_batch_begin": False,
            "on_train_loss_begin": False,
            "on_backward_begin": False,
            "on_backward_end": False,
            "on_train_step_end": False,
            "on_train_batch_end": False,
            "on_validation_begin": False,
            "on_validation_epoch_begin": False,
            "on_validation_batch_begin": False,
            "on_validation_loss_begin": False,
            "on_validation_step_end": False,
            "on_validation_batch_end": False,
            "on_validation_epoch_end": False,
            "on_validation_end": False,
            "on_train_epoch_end": False,
            "on_train_end": False,
            "on_test_begin": False,
            "on_test_epoch_begin": False,
            "on_test_batch_begin": False,
            "on_test_loss_begin": False,
            "on_test_step_end": False,
            "on_test_batch_end": False,
            "on_test_epoch_end": False,
            "on_test_end": False,
            "on_teardown": False,
        }

    def on_setup(self, context):
        self.method_is_called["on_setup"] = True

    def on_init_begin(self, context):
        self.method_is_called["on_init_begin"] = True

    def on_init_end(self, context):
        self.method_is_called["on_init_end"] = True

    def on_train_begin(self, context):
        self.method_is_called["on_train_begin"] = True

    def on_train_epoch_begin(self, context):
        self.method_is_called["on_train_epoch_begin"] = True

    def on_train_batch_begin(self, context):
        self.method_is_called["on_train_batch_begin"] = True

    def on_train_loss_begin(self, context):
        self.method_is_called["on_train_loss_begin"] = True

    def on_backward_begin(self, context):
        self.method_is_called["on_backward_begin"] = True

    def on_backward_end(self, context):
        self.method_is_called["on_backward_end"] = True

    def on_train_step_end(self, context):
        self.method_is_called["on_train_step_end"] = True

    def on_train_batch_end(self, context):
        self.method_is_called["on_train_batch_end"] = True

    def on_validation_begin(self, context):
        self.method_is_called["on_validation_begin"] = True

    def on_validation_epoch_begin(self, context):
        self.method_is_called["on_validation_epoch_begin"] = True

    def on_validation_batch_begin(self, context):
        self.method_is_called["on_validation_batch_begin"] = True

    def on_validation_loss_begin(self, context):
        self.method_is_called["on_validation_loss_begin"] = True

    def on_validation_step_end(self, context):
        self.method_is_called["on_validation_step_end"] = True

    def on_validation_batch_end(self, context):
        self.method_is_called["on_validation_batch_end"] = True

    def on_validation_epoch_end(self, context):
        self.method_is_called["on_validation_epoch_end"] = True

    def on_validation_end(self, context):
        self.method_is_called["on_validation_end"] = True

    def on_train_epoch_end(self, context):
        self.method_is_called["on_train_epoch_end"] = True

    def on_train_end(self, context):
        self.method_is_called["on_train_end"] = True

    def on_test_begin(self, context):
        self.method_is_called["on_test_begin"] = True

    def on_test_epoch_begin(self, context):
        self.method_is_called["on_test_epoch_begin"] = True

    def on_test_batch_begin(self, context):
        self.method_is_called["on_test_batch_begin"] = True

    def on_test_loss_begin(self, context):
        self.method_is_called["on_test_loss_begin"] = True

    def on_test_step_end(self, context):
        self.method_is_called["on_test_step_end"] = True

    def on_test_batch_end(self, context):
        self.method_is_called["on_test_batch_end"] = True

    def on_test_epoch_end(self, context):
        self.method_is_called["on_test_epoch_end"] = True

    def on_test_end(self, context):
        self.method_is_called["on_test_end"] = True

    def on_teardown(self, context):
        self.method_is_called["on_teardown"] = True


CallbackMethodIsCalled = CallbackMethodIsCalledRecorder



import click
import mlflow
from mlflow.pyfunc.scoring_server import init

from roheboam.engine import get_toolbox_lookup
from roheboam.engine.integrations.mlflow.saving import save_mlflow_pyfunc_model
from roheboam.engine.pipeline import Pipeline
from roheboam.engine.pipeline.config_loader import load_config_from_path
from roheboam.engine.pipeline.constants import VARIABLE_KEY
from roheboam.engine.vision.deployment import run_seldon_image_model_server


@click.group()
def cli():
    pass


# Before are the commands that the server will call indirectly through
# Docker CLI through the endpoint defined by the application hosted
@click.command()
@click.option("--config_path")
@click.option("--root_save_directory")
@click.option("--gpus", default=None)
def train(config_path, root_save_directory, gpus):

    config = load_config_from_path(config_path)
    config[VARIABLE_KEY]["root_save_directory"] = root_save_directory
    config[VARIABLE_KEY]["gpus"] = gpus
    lookup = get_toolbox_lookup()
    pipeline = Pipeline.create_from_config(config, lookup)
    pipeline.run()

    train_sample_idx, val_sample_idx = next(
        iter(pipeline.get_node_output("DataSplitCreator")().split())
    )
    try:
        train_samples = pipeline.get_node_output("TrainSampleDataCreator")[
            train_sample_idx
        ]
    except KeyError:
        train_samples = pipeline.get_node_output("TrainSamples")[train_sample_idx]

    try:
        val_samples = pipeline.get_node_output("TrainSampleDataCreator")[val_sample_idx]
    except KeyError:
        val_samples = pipeline.get_node_output("TrainSamples")[val_sample_idx]

    try:
        test_samples = pipeline.get_node_output("TestSampleDataCreator")
    except KeyError:
        test_samples = pipeline.get_node_output("TestSamples")

    module = pipeline.get_node_output("ModuleCreator")(
        train_samples=train_samples, val_samples=val_samples, test_samples=test_samples
    )

    checkpoint_callback = pipeline.get_node_output("PyTorchLightningModelCheckpoint")
    trainer = pipeline.get_node_output("TrainerCreator")(
        checkpoint_callback=checkpoint_callback
    )
    trainer.fit(module)

    mlflow_save_path = pipeline.get_node_output("TimeStampedSavePath") / "mlflow"
    save_mlflow_pyfunc_model(
        mlflow_save_path,
        pipeline.get_node_output("ServingModel"),
        artifacts={
            "config_path": str(checkpoint_callback.best_config_path),
            "model_save_path": str(checkpoint_callback.best_model_path),
        },
    )


@click.command()
@click.option("--model_path")
@click.option("--model_format")
@click.option("--debug", is_flag=True)
def serve(model_path, model_format, debug):
    assert model_path is not None, "No model path is specified"
    if model_format == "mlflow":
        model = mlflow.pyfunc.load_model(str(model_path))
        app = init(model)
        app.run(host="0.0.0.0", debug=debug)

    if model_format == "seldon":
        with run_seldon_image_model_server(model_path, debug):
            pass


cli.add_command(train)
cli.add_command(serve)

if __name__ == "__main__":
    cli()



from itertools import repeat

import numpy as np
from p_tqdm import p_map

from ..data import create_image_segmentation_sample


def create_image_segmentation_samples_data_from_paths(
    image_paths, mask_paths=None, weight_map_paths=None
):
    stringify = lambda l: [str(e) for e in l]
    image_paths = stringify(image_paths)
    mask_paths = repeat(None) if mask_paths is None else stringify(mask_paths)
    weight_map_paths = (
        repeat(None) if weight_map_paths is None else stringify(weight_map_paths)
    )
    return np.array(
        [
            {
                "image_path": image_path,
                "mask_path": mask_path,
                "weight_map_path": weight_map_path,
            }
            for image_path, mask_path, weight_map_path in zip(
                image_paths, mask_paths, weight_map_paths
            )
        ]
    )


def create_image_segmentation_samples_from_paths(
    image_paths, mask_paths=None, weight_map_paths=None
):
    samples = p_map(
        _create_image_segmentation_sample_lambda,
        create_image_segmentation_samples_data_from_paths(
            image_paths, mask_paths, weight_map_paths
        ),
    )
    return np.array(samples)


def _create_image_segmentation_sample_lambda(sample_data):
    return create_image_segmentation_sample(**sample_data)


lookup = {
    "create_image_segmentation_samples_data_from_paths": create_image_segmentation_samples_data_from_paths,
    "create_image_segmentation_samples_from_paths": create_image_segmentation_samples_from_paths,
}



import os
import traceback
from contextlib import ContextDecorator
from typing import Dict, Iterable, List, Optional, Union

import numpy as np

from ...logger import logger


class run_seldon_image_model_server(ContextDecorator):
    def __init__(self, model_path, debug=False, background_process=False):
        self.model_path = model_path
        self.debug = debug
        self.background_process = background_process

    def __enter__(self):
        cmd = "seldon-core-microservice roheboam.engine.vision.SeldonImageModel --service-type MODEL"
        if self.debug:
            cmd += " --debug"
        os.environ["SELDON_MODEL_PATH"] = self.model_path
        from ...utils.convenience import run_shell_command

        run_shell_command(cmd, background_process=self.background_process)

    def __exit__(self, *exc):
        print("Exiting")
        del os.environ["SELDON_MODEL_PATH"]


class SeldonImageModel:
    def __init__(self):
        self.loaded = False
        pass

    def load(self):
        import mlflow
        import numpy as np
        from pytorch_lightning.utilities.cloud_io import load as pl_load

        from ....engine import get_toolbox_lookup

        mlflow_model = mlflow.pyfunc.load_model(
            os.environ.get("SELDON_MODEL_PATH", str("/model"))
        )
        context = mlflow_model._model_impl.context

        self.lookup = get_toolbox_lookup()
        self.config = self.lookup["load_config_from_path"](
            context.artifacts["config_path"]
        )
        self.config["Variables"]["is_deployed"] = True
        self.pipeline = self.lookup["Pipeline"].create_from_config(
            self.config, self.lookup
        )
        self.pipeline.run(remove_nodes_with_tags=["TRAIN"])
        self.module = self.pipeline.get_node_output("ModuleCreator")()
        self.trainer = self.pipeline.get_node_output("TrainerCreator")()
        self.inference_recorder = self.pipeline.get_node_output("InferenceRecorder")
        pl_checkpoint = pl_load(
            context.artifacts["model_save_path"],
            map_location=lambda storage, loc: storage,
        )
        self.module.load_state_dict(pl_checkpoint["state_dict"])
        self.loaded = True

    def predict(
        self, X: np.ndarray, names: Optional[List[str]], meta: Dict = None
    ) -> Union[np.ndarray, List, str, bytes]:
        try:
            samples_data = [
                {"image_data": self.lookup["to_uint8_image"](image)} for image in X
            ]
            samples = [
                self.pipeline.get_node_output("SampleFromDataCreator")(**sample_data)
                for sample_data in samples_data
            ]
            test_dataset = self.pipeline.get_node_output("TestDatasetCreator")(
                samples=samples
            )
            test_dataloader = self.pipeline.get_node_output("TestDataLoaderCreator")(
                test_dataset
            )
            self.trainer.test(self.module, dataloaders=[test_dataloader])
            predictions = self.inference_recorder.serializable_results
            self.inference_recorder.flush_results()
            return predictions
        except Exception:
            logger.error(traceback.format_exc())

    def health_status(self):
        assert self.loaded, "Model is not loaded"
        return ["Success"]


lookup = {
    "SeldonImageModel": SeldonImageModel,
    "run_seldon_image_model_server": run_seldon_image_model_server,
}



from copy import deepcopy

import cv2
import matplotlib.pyplot as plt
import numpy as np


def plot_image_grid(images, titles=None, ncols=None, cmap="gray", figsize=None):
    if titles is not None:
        assert len(images) == len(titles)
    if not ncols:
        factors = [i for i in range(1, len(images) + 1) if len(images) % i == 0]
        ncols = factors[len(factors) // 2] if len(factors) else len(images) // 4 + 1
    nrows = int(len(images) / ncols) + int(len(images) % ncols)
    imgs = [images[i] if len(images) > i else None for i in range(nrows * ncols)]

    if figsize is None:
        figsize = (3 * ncols, 2 * nrows)
    fig, axes = plt.subplots(nrows, ncols, figsize=figsize)
    axes = axes.flatten()[: len(imgs)]
    for i, (img, ax) in enumerate(zip(imgs, axes.flatten())):
        if len(img.shape) > 2 and img.shape[2] == 1:
            img = img.squeeze()
        ax.imshow(img, cmap=cmap)
        if titles is not None:
            ax.set_title(titles[i])
    fig.tight_layout()


def plot_polygons_on_image(
    image, polygons, is_closed=True, color=(0, 255, 0), thickness=1
):
    """Plot bbox on an image

    :param image: np.array with H x W x C
    :type image: list
    :param bboxes: list of polygons in the OpenCV format: [[x1, y1], [x2, y2], ...]
    :type bboxes: list
    :param color: [description], defaults to (0, 255, 0)
    :type color: tuple, optional
    :param thickness: [description], defaults to 1
    :type thickness: int, optional
    """
    from . import is_rgb

    assert is_rgb(image), "Please make sure that the image is RGB"

    image = deepcopy(image)
    for polygon in [polygons]:
        cv2.polylines(image, [np.array(polygon)], is_closed, color, thickness)
    return image


def plot_bounding_boxes_on_image_with_label(
    image, bboxes, labels, label_bbox_metas=None
):
    """Plot bbox with labels on image

    :param image: np.array with H x W x C
    :type image: list
    :param bboxes: list of bbox in the pascal_format: [x_min, y_min, x_max, y_max]
    :type bboxes: list
    :param labels: list of labels for each bbox
    :type labels: list
    :param label_bbox_metas: list of dicts {label: <int>, color: (r, g, b), thickness: <int>}
    :type label_bbox_metas: list
    """
    labels = [int(label) for label in deepcopy(labels)]

    if label_bbox_metas is None:
        rgb_map = _generate_default_rgb_map()
        label_bbox_metas = [
            {
                "label": label,
                "name": str(label),
                "color": rgb_map[label % len(rgb_map)],
                "thickness": 3,
            }
            for label in labels
        ]

    # deepcopy and avoid https://github.com/opencv/opencv/issues/18120
    image = np.ones(image.shape, np.uint8) * image

    for bbox, label_to_plot in zip(bboxes, labels):
        x_min, y_min, x_max, y_max = [int(f) for f in bbox]
        label_bbox_meta = [
            label for label in label_bbox_metas if label["label"] == label_to_plot
        ][0]
        cv2.rectangle(
            image,
            (x_min, y_min),
            (x_max, y_max),
            label_bbox_meta["color"],
            label_bbox_meta["thickness"],
        )
        cv2.putText(
            image,
            label_bbox_meta["name"],
            (x_min, y_min - 2),
            0,
            1,
            [225, 255, 255],
            thickness=label_bbox_meta["thickness"],
            lineType=cv2.LINE_AA,
        )
    #         cv2.putText(img, label, (c1[0], c1[1] - 2), 0, tl / 3, [225, 255, 255], thickness=tf, lineType=cv2.LINE_AA)
    return image


def _generate_default_rgb_map():
    prop_cycle = plt.rcParams["axes.prop_cycle"]
    # https://stackoverflow.com/questions/51350872/python-from-color-name-to-rgb
    hex2rgb = lambda h: tuple(int(h[1 + i : 1 + i + 2], 16) for i in (0, 2, 4))
    return [hex2rgb(h) for h in prop_cycle.by_key()["color"]]


def plot_bounding_boxes_on_image(image, bboxes, color=(0, 255, 0), thickness=1):
    """Plot bbox on an image

    :param image: np.array with H x W x C
    :type image: list
    :param bboxes: list of bbox in the pascal_format: [x_min, y_min, x_max, y_max]
    :type bboxes: list
    :param color: [description], defaults to (0, 255, 0)
    :type color: tuple, optional
    :param thickness: [description], defaults to 1
    :type thickness: int, optional
    """

    from . import is_rgb

    assert is_rgb(image), "Please make sure that the image is RGB"

    # deepcopy and avoid https://github.com/opencv/opencv/issues/18120
    image = np.ones(image.shape, np.uint8) * image
    for bbox in bboxes:
        x_min, y_min, x_max, y_max = bbox
        cv2.rectangle(image, (x_min, y_min), (x_max, y_max), color, thickness)
    return image


def plot_mask_on_image(image, mask, image_alpha=0.2):
    from . import to_rgb, to_uint8_image

    return cv2.addWeighted(
        to_uint8_image(image),
        image_alpha,
        to_uint8_image(to_rgb(mask)),
        1 - image_alpha,
        0,
    )


lookup = {
    "plot_image_grid": plot_image_grid,
    "plot_polygons_on_image": plot_polygons_on_image,
    "plot_bounding_boxes_on_image_with_label": plot_bounding_boxes_on_image_with_label,
    "plot_bounding_boxes_on_image": plot_bounding_boxes_on_image,
    "plot_mask_on_image": plot_mask_on_image,
}



from pathlib import Path

from ...utils.convenience import run_shell_command


class DVC:
    def __init__(self, dvc_run_path=None, silent=False, construct_only=False):
        self.default_dvc_run_path = dvc_run_path
        self.default_silent = silent
        self.construct_only = construct_only

    def _run_shell_command_with_log_status(self, cmd, silent):
        cmd = " ".join(cmd)
        if self.construct_only:
            print(cmd)
        else:
            run_shell_command(
                cmd,
                silent=self.default_silent if silent is None else silent,
                same_process=False,
            )

    def _construct_dvc_command_with_run_path(self, dvc_run_path):
        cmd = ["dvc"]
        dvc_run_path = (
            self.default_dvc_run_path if dvc_run_path is None else dvc_run_path
        )
        if dvc_run_path:
            cmd.extend(["--cd", str(dvc_run_path)])
        return cmd

    def init(
        self, sub_dir=None, force=False, verbose=False, dvc_run_path=None, silent=False
    ):
        cmd = self._construct_dvc_command_with_run_path(dvc_run_path)
        cmd.extend(["init"])

        if sub_dir:
            cmd.extend(["--subdir"])
        if force:
            cmd.extend(["--force"])
        if verbose:
            cmd.extend(["--verbose"])
        self._run_shell_command_with_log_status(cmd, silent)

    def install(self, dvc_run_path=None, silent=False):
        cmd = self._construct_dvc_command_with_run_path(dvc_run_path)
        cmd.extend(["install"])
        self._run_shell_command_with_log_status(cmd, silent)

    def add(self, add_path, dvc_output_path=None, dvc_run_path=None, silent=False):
        cmd = self._construct_dvc_command_with_run_path(dvc_run_path)
        cmd.extend(["add", str(add_path)])
        if dvc_output_path is not None:
            Path(dvc_output_path).parent.mkdir(exist_ok=True, parents=True)
            cmd.extend(["--file", str(dvc_output_path)])
            self._run_shell_command_with_log_status(cmd, silent)

    def add_remote(
        self, remote_name, remote_path, as_default=True, dvc_run_path=None, silent=False
    ):
        cmd = self._construct_dvc_command_with_run_path(dvc_run_path)
        cmd.extend(["remote", "add"])

        if as_default:
            cmd.extend(["-d"])
        cmd.extend([remote_name, remote_path])
        self._run_shell_command_with_log_status(cmd, silent)

    def modify_remote_endpoint(
        self, remote_name, endpoint_url, dvc_run_path=None, silent=False
    ):
        cmd = self._construct_dvc_command_with_run_path(dvc_run_path)
        cmd.extend(
            ["remote", "modify", "--local", remote_name, "endpointurl", endpoint_url]
        )
        self._run_shell_command_with_log_status(cmd, silent)

    def modify_remote_access_keys_id(
        self, remote_name, access_key_id, dvc_run_path=None, silent=False
    ):
        cmd = self._construct_dvc_command_with_run_path(dvc_run_path)
        cmd.extend(
            ["remote", "modify", "--local", remote_name, "access_key_id", access_key_id]
        )
        self._run_shell_command_with_log_status(cmd, silent)

    def modify_remote_secret_access_key(
        self, remote_name, secret_access_key, dvc_run_path=None, silent=False
    ):
        cmd = self._construct_dvc_command_with_run_path(dvc_run_path)
        cmd.extend(
            [
                "remote",
                "modify",
                "--local",
                remote_name,
                "secret_access_key",
                secret_access_key,
            ]
        )
        self._run_shell_command_with_log_status(cmd, silent)

    def run(
        self,
        name,
        run_command,
        dependency_paths=[],
        output_paths=[],
        dvc_run_path=None,
        force=False,
        no_run_cache=False,
        silent=False,
    ):
        cmd = self._construct_dvc_command_with_run_path(dvc_run_path)
        cmd.extend(["run"])
        cmd.extend(["-n", name])
        if force:
            cmd.extend(["--force"])
        if no_run_cache:
            cmd.extend(["--no-run-cache"])
        cmd.extend([f"-d {str(p)}" for p in dependency_paths])
        cmd.extend([f"-o {str(p)}" for p in output_paths])
        cmd.extend([run_command])
        self._run_shell_command_with_log_status(cmd, silent)

    def repro(self, target=None, dvc_run_path=None, silent=False):
        cmd = self._construct_dvc_command_with_run_path(dvc_run_path)
        if target is not None:
            cmd.extend([target])
        cmd.extend(["repro"])
        self._run_shell_command_with_log_status(cmd, silent)

    def pull(self, stage=None, dvc_run_path=None, silent=False):
        cmd = self._construct_dvc_command_with_run_path(dvc_run_path)
        cmd.extend(["pull"])
        if stage:
            cmd.extend([stage])
        self._run_shell_command_with_log_status(cmd, silent)

    def push(self, dvc_run_path=None, silent=False):
        cmd = self._construct_dvc_command_with_run_path(dvc_run_path)
        cmd.extend(["push"])
        self._run_shell_command_with_log_status(cmd, silent)



from pathlib import Path

import cloudpickle
import mlflow
from pytorch_lightning.utilities import rank_zero_only

from ....utils.conda import get_current_conda_env
from ....utils.saving import save_to_temporary_file
from .. import Callback


class MLflowModelSaver(Callback):
    def __init__(
        self,
        config,
        lookup,
        pipeline,
        mlflow_model,
        root_save_path,
        conda_env_name=None,
        lookup_save_name="lookup.p",
    ):
        self.config = config
        self.lookup = lookup
        self.lookup_save_name = lookup_save_name
        self.mlflow_model = mlflow_model
        self.mlflow_model_save_path = Path(root_save_path) / "mlflow"
        self.conda_env_name = conda_env_name

    def save_lookup(self, file_name):
        cloudpickle.dump(self.lookup, open(file_name, "wb"))

    @rank_zero_only
    def on_teardown(self, context):
        with save_to_temporary_file(
            self.save_lookup, self.lookup_save_name
        ) as lookup_file_path:
            conda_env = get_current_conda_env(env_name=self.conda_env_name)
            artifacts = {
                "config_path": str(self.config.save_path),
                "model_save_path": str(self.config.model_save_path),
                "lookup_path": str(lookup_file_path),
            }
            mlflow.pyfunc.save_model(
                path=self.mlflow_model_save_path,
                python_model=self.mlflow_model,
                artifacts=artifacts,
                conda_env=conda_env,
            )



from .conda import lookup as conda_lookup
from .convenience import lookup as convenience_lookup
from .hooks import lookup as hooks_lookup
from .interfaces import lookup as interfaces_lookup
from .model import lookup as models_lookup
from .saving import lookup as saving_lookup
from .testing import lookup as testing_lookup
from .training import lookup as training_lookup

lookup = {
    **conda_lookup,
    **convenience_lookup,
    **hooks_lookup,
    **interfaces_lookup,
    **models_lookup,
    **saving_lookup,
    **testing_lookup,
    **training_lookup,
}



from collections import defaultdict
from copy import deepcopy

import numpy as np

from ....utils.convenience import to_numpy
from ....vision.utils.bbox import nms_with_scores
from ....vision.utils.image import plot_image_grid
from ..callback import Callback
from .utils import ShouldRecordMixin


class _PlotMixins:
    def plot_sample_with_thresholds(
        self, sample_with_thresholds, ncols=None, figsize=None
    ):
        images, titles = self.create_images_and_titles_from_sample_with_thresholds(
            sample_with_thresholds
        )
        plot_image_grid(images, titles=titles, ncols=ncols, figsize=figsize)

    def create_images_and_titles_from_sample_with_thresholds(
        self, sample_with_thresholds
    ):
        images = []
        titles = []
        for threshold, sample in sample_with_thresholds.items():
            images.append(sample.create_plot_array())
            titles.append(threshold)
        return images, titles


class _InferenceMixins:
    def create_samples_with_thresholds_from_samples_with_pipeline(
        self, samples, pipeline, trainer, module, thresholds=np.arange(0, 1, 0.2)
    ):
        dataloader = pipeline.get_node_output("TestDataLoaderCreator")(
            pipeline.get_node_output("TestDatasetCreator")(samples)
        )
        results = self.create_results_from_dataloader(dataloader, trainer, module)
        return [
            self.create_samples_with_thresholds_from_result(
                result, sample.bboxes_format, thresholds
            )
            for sample, result in zip(samples, results)
        ]

    def create_samples_with_thresholds_from_samples_data_with_pipeline(
        self, samples_data, pipeline, trainer, module, thresholds=np.arange(0, 1, 0.2)
    ):
        dataloader = pipeline.get_node_output("TestDataLoaderCreator")(
            pipeline.get_node_output("TestDatasetCreator")(samples_data)
        )
        results = self.create_results_from_dataloader(dataloader, trainer, module)
        return [
            self.create_samples_with_thresholds_from_result(result, thresholds)
            for result in results
        ]

    def create_results_from_dataloader(self, dataloader, trainer, module):
        trainer.test(module, dataloaders=[dataloader])
        results = deepcopy(self.results)
        self.flush_results()
        return results

    def create_samples_with_thresholds_from_result(
        self, result, bboxes_format, thresholds=np.arange(0, 1, 0.2), include_truth=True
    ):
        sample_with_thresholds = {
            str(np.around(threshold, 2)): self.create_predicted_sample(
                result, bboxes_format, threshold
            )
            for threshold in thresholds
        }

        if include_truth and "truth_labels" in result and "truth_boxes" in result:
            sample_with_thresholds["truth"] = self.create_truth_sample(
                result, bboxes_format
            )
        return sample_with_thresholds

    def create_predicted_and_truth_samples(
        self, result, bboxes_format, iou_threshold=0.2
    ):
        predicted_sample = self.create_predicted_sample(
            result, bboxes_format, iou_threshold
        )
        truth_sample = self.create_truth_sample(result, bboxes_format)
        return predicted_sample, truth_sample

    def create_predicted_sample(self, result, bboxes_format, iou_threshold=0.2):
        from ....vision.tasks.image_bounding_box_detection.data import ImageBoundingBoxDetectionSample

        nms_outputs = nms_with_scores(
            result["predicted_boxes"],
            result["predicted_scores"],
            result["predicted_labels"],
            iou_threshold=iou_threshold,
        )
        predicted_sample = ImageBoundingBoxDetectionSample.create(
            image_data=result["image"],
            labels_data=nms_outputs["labels"],
            bboxes_data=nms_outputs["boxes"],
            bboxes_format=bboxes_format,
        )
        return predicted_sample

    def create_truth_sample(self, result, bboxes_format):
        from ....vision.tasks.image_bounding_box_detection.data import ImageBoundingBoxDetectionSample

        truth_sample = ImageBoundingBoxDetectionSample.create(
            image_data=result["image"],
            labels_data=result["truth_labels"],
            bboxes_data=result["truth_boxes"],
            bboxes_format=bboxes_format,
        )
        return truth_sample


class ImageBoundingBoxDetectionResultRecorder(
    Callback, ShouldRecordMixin, _InferenceMixins, _PlotMixins
):

    _order = -10

    def __init__(
        self,
        convert_tensor_to_image_fn=lambda x: x,
        convert_tensor_to_bboxes_fn=lambda x: x,
        convert_tensor_to_labels_fn=to_numpy,
        convert_tensor_to_scores_fn=to_numpy,
        record_on_validation=True,
        record_on_test=True,
        record_every_n_steps=25,
        name=None,
    ):
        super().__init__(name)
        self.convert_tensor_to_image_fn = convert_tensor_to_image_fn
        self.convert_tensor_to_bboxes_fn = convert_tensor_to_bboxes_fn
        self.convert_tensor_to_labels_fn = convert_tensor_to_labels_fn
        self.convert_tensor_to_scores_fn = convert_tensor_to_scores_fn
        self.record_on_validation = record_on_validation
        self.record_on_test = record_on_test
        self.record_every_n_steps = record_every_n_steps
        self.results = defaultdict(list)

    def on_train_batch_end(self, context):

        phase, epoch, step, batch, _ = (
            context.current_phase,
            context.current_epoch,
            context.current_step,
            context.train_batch,
            context.train_prediction,
        )

        if not self._should_store(phase, epoch, step):
            return

        self.results[(str(phase), epoch)].extend(
            [
                {"image": self.convert_tensor_to_image_fn(image)}
                for image in batch["images"]
            ]
        )

    def on_validation_batch_end(self, context):
        phase, epoch, step, batch, predictions = (
            context.current_phase,
            context.current_epoch,
            context.current_step,
            context.validation_batch,
            context.validation_prediction,
        )
        if not self._should_store(phase, epoch, step):
            return

        self.results[(str(phase), epoch)].extend(
            [
                {
                    "image": self.convert_tensor_to_image_fn(image),
                    "predicted_boxes": self.convert_tensor_to_bboxes_fn(
                        prediction["boxes"]
                    ),
                    "predicted_scores": self.convert_tensor_to_scores_fn(
                        prediction["scores"]
                    ),
                    "predicted_labels": self.convert_tensor_to_labels_fn(
                        prediction["labels"]
                    ),
                    "truth_boxes": self.convert_tensor_to_bboxes_fn(target["boxes"]),
                    "truth_labels": self.convert_tensor_to_labels_fn(target["labels"]),
                }
                for image, target, prediction in zip(
                    batch["images"], batch["targets"], predictions
                )
            ]
        )

    def on_test_batch_end(self, context):
        phase, epoch, step, batch, predictions = (
            context.current_phase,
            context.current_epoch,
            context.current_step,
            context.test_batch,
            context.test_prediction,
        )
        if not self._should_store(phase, epoch, step):
            return

        self.results[(str(phase), epoch)].extend(
            [
                {
                    "image": self.convert_tensor_to_image_fn(image),
                    "predicted_boxes": self.convert_tensor_to_bboxes_fn(
                        prediction["boxes"]
                    ),
                    "predicted_scores": self.convert_tensor_to_scores_fn(
                        prediction["scores"]
                    ),
                    "predicted_labels": self.convert_tensor_to_labels_fn(
                        prediction["labels"]
                    ),
                    "truth_boxes": self.convert_tensor_to_bboxes_fn(target["boxes"]),
                    "truth_labels": self.convert_tensor_to_labels_fn(target["labels"]),
                }
                for image, target, prediction in zip(
                    batch["images"], batch["targets"], predictions
                )
            ]
        )


class ImageBoundingBoxDetectionInferenceRecorder(
    Callback, _InferenceMixins, _PlotMixins
):
    _order = -10

    def __init__(
        self,
        convert_tensor_to_image_fn=lambda x: x,
        convert_tensor_to_bboxes_fn=lambda x: x,
        convert_tensor_to_labels_fn=to_numpy,
        convert_tensor_to_scores_fn=to_numpy,
        name=None,
    ):
        super().__init__(name)
        self.convert_tensor_to_image_fn = convert_tensor_to_image_fn
        self.convert_tensor_to_bboxes_fn = convert_tensor_to_bboxes_fn
        self.convert_tensor_to_labels_fn = convert_tensor_to_labels_fn
        self.convert_tensor_to_scores_fn = convert_tensor_to_scores_fn
        self.results = []

    @property
    def serializable_results(self):
        serializable_results = []
        for result in self.results:
            serializable_results.append({k: v.tolist() for k, v in result.items()})
        return serializable_results

    def flush_results(self):
        self.results = []

    def on_test_batch_end(self, context):
        batch, predictions = (context.test_batch, context.test_prediction)

        if "targets" in batch:
            self.results.extend(
                [
                    {
                        "image": self.convert_tensor_to_image_fn(image),
                        "predicted_boxes": self.convert_tensor_to_bboxes_fn(
                            prediction["boxes"]
                        ),
                        "predicted_scores": self.convert_tensor_to_scores_fn(
                            prediction["scores"]
                        ),
                        "predicted_labels": self.convert_tensor_to_labels_fn(
                            prediction["labels"]
                        ),
                        "truth_boxes": self.convert_tensor_to_bboxes_fn(
                            target["boxes"]
                        ),
                        "truth_labels": self.convert_tensor_to_labels_fn(
                            target["labels"]
                        ),
                    }
                    for image, target, prediction in zip(
                        batch["images"], batch["targets"], predictions
                    )
                ]
            )
        else:
            self.results.extend(
                [
                    {
                        "image": self.convert_tensor_to_image_fn(image),
                        "predicted_boxes": self.convert_tensor_to_bboxes_fn(
                            prediction["boxes"]
                        ),
                        "predicted_scores": self.convert_tensor_to_scores_fn(
                            prediction["scores"]
                        ),
                        "predicted_labels": self.convert_tensor_to_labels_fn(
                            prediction["labels"]
                        ),
                    }
                    for image, prediction in zip(batch["images"], predictions)
                ]
            )



from torchvision.ops import box_iou as torchvision_box_iou

from ...utils.training import tensorify
from .core import BaseMetric


class BoxIOU(BaseMetric):
    def __init__(self, threshold=0.5, name=None):
        super().__init__(name)
        self.threshold = threshold

    @property
    def unreduced_metric(self):
        return self._unreduced_metric

    @property
    def per_sample_metric(self):
        return self.metric

    @property
    def reduced_metric(self):
        return self.metric

    @property
    def total_metric(self):
        return self.metric

    def to_call_kwargs(self, d):
        predicted_boxes = d["predictions"]["boxes"]
        predicted_labels = d["predictions"]["labels"]
        predicted_scores = d["predictions"]["scores"]
        target_boxes = d["targets"]["boxes"]
        target_labels = d["targets"]["labels"]
        return {
            "predicted_boxes": predicted_boxes,
            "predicted_labels": predicted_labels,
            "predicted_scores": predicted_scores,
            "target_boxes": target_boxes,
            "target_labels": target_labels,
        }

    def __call__(
        self,
        predicted_boxes,
        predicted_labels,
        predicted_scores,
        target_boxes,
        target_labels,
    ):

        self._unreduced_metric = box_iou(predicted_boxes, target_boxes)
        self._per_sample_metric = self._unreduced_metric.diag()
        self._reduced_metric = self._per_sample_metric.mean()
        self._total_metric = self._reduced_metric
        return self._total_metric


def box_iou(predictions, targets):
    """Return intersection-over-union (Jaccard index) of boxes.

    Args:
        prediction (N, 4): ground truth boxes.
        target (N, 4): predicted boxes.
        eps: a small number to avoid 0 as denominator.
    Returns:
        iou (N): IoU values.
    """
    return torchvision_box_iou(
        tensorify(predictions).float(), tensorify(targets).float()
    )


lookup = {"BoxIOU": BoxIOU, "box_iou": box_iou}



import inspect
from copy import deepcopy
from pprint import pformat

from ..logger import logger
from .yaml_loader import Reference


def flatten_resources_dict(d):
    flattened = dict()
    for k, v in d.items():
        assert (
            type(v) is not str
        ), f"The field: {v} is not a dictionary\nCheck if there is a field named 'properties' in the in your node containing fields:\n{pformat(d)}"

        if "properties" in v.keys():
            flattened.update(**{k: v})
        else:
            flattened.update(**flatten_resources_dict(v))
    return flattened


def find_references(resource):
    references = []
    if isinstance(resource, dict):
        for _, value in resource.items():
            references.extend(find_references(value))
    elif isinstance(resource, list):
        for value in resource:
            references.extend(find_references(value))
    elif isinstance(resource, Reference):
        references.append(resource)
    else:
        pass
    return references


def replace_arguments(pipeline, node):
    try:
        arguments = node.arguments
        needs_pipeline = "pipeline" in inspect.signature(node.pointer).parameters
        needs_config = "config" in inspect.signature(node.pointer).parameters
        needs_lookup = "lookup" in inspect.signature(node.pointer).parameters

        if arguments is not None:
            reference_replaced_arguments = replace_references(
                pipeline.graph, deepcopy(arguments)
            )
            if needs_pipeline:
                reference_replaced_arguments["pipeline"] = pipeline
            if needs_config:
                reference_replaced_arguments["config"] = pipeline.config
            if needs_lookup:
                reference_replaced_arguments["lookup"] = pipeline.lookup

            return reference_replaced_arguments
    except AttributeError as e:
        logger.warning(e)
        return {}


def replace_should_remove(graph, node):
    return replace_references(graph, deepcopy(node.should_remove))


def replace_should_run(graph, node):
    return replace_references(graph, deepcopy(node.should_run))


def replace_references(graph, references):
    if isinstance(references, dict):
        for name, reference in references.items():
            references[name] = replace_references(graph, reference)
    elif isinstance(references, list):
        for i, reference in enumerate(references):
            references[i] = replace_references(graph, reference)
    elif isinstance(references, Reference):
        # reassign name it make the intent clearer
        reference = references
        ref_node = graph.nodes(data=True)[reference.ref_node_name]["node"]
        ref_node_outputs = ref_node.output
        assert (
            ref_node_outputs is not None
        ), f"Node: {reference.ref_node_name} has no output"
        assert (
            reference.output_name in ref_node_outputs
        ), f"Node: {reference.ref_node_name} has no output named {reference.output_name}"
        references = ref_node_outputs[reference.output_name]
    else:
        return references
    return references



from .callbacks import lookup as callbacks_lookup
from .data import lookup as data_lookup
from .losses import lookup as losses_lookup
from .metrics import lookup as metrics_lookup
from .optimizer import lookup as optimizer_lookup
from .scheduler import lookup as scheduler_lookup
from .training import lookup as training_lookup

lookup = {
    **callbacks_lookup,
    **data_lookup,
    **losses_lookup,
    **metrics_lookup,
    **optimizer_lookup,
    **scheduler_lookup,
    **training_lookup,
}



from collections import defaultdict

from ..handler import Phase


class Storage:
    def __init__(self):
        self._storage = defaultdict(list)

    def read(self, phase, epoch, key):
        return self._storage[(phase, epoch, key)]

    def write(self, phase, epoch, key, data=[]):
        assert isinstance(data, list) or isinstance(data, tuple)
        self._storage[(phase, epoch, key)].extend(data)


class ShouldRecordMixin:
    def _should_store(self, phase, epoch, step):
        if step == 0 or step % self.record_every_n_steps == 0:
            return True
        if phase == Phase.VALIDATION and self.record_on_validation:
            return True
        if phase == Phase.TEST and self.record_on_test:
            return True
        return False



import importlib.util
import shutil
import tempfile
from pathlib import Path
from textwrap import dedent

from ...logger import logger
from ..docker.utils import build_image


def containerise_model(
    model_path,
    model_format,
    model_image_name,
    build_env_image,
    dev_build=False,
    conda_env_file_path=None,
):
    if dev_build and (conda_env_file_path is None):
        raise ValueError(
            "There must be a conda_lock_file_path and poetry_lock_file_path if dev_build is enabled"
        )

    with tempfile.TemporaryDirectory() as temp_dir:
        if dev_build:
            logger.info("Copying roheboam source to build directory")
            _copy_roheboam(temp_dir)
            _copy_dependency_files(temp_dir, conda_env_file_path)

        logger.info(f"Copying model from {model_path} to build directory")
        _copy_model(model_path, temp_dir)

        logger.info(f"Creating Docker file")
        docker_file = _create_docker_file(
            model_path, model_format, build_env_image, dev_build
        )

        logger.info(f"Writing Docker file")
        _write_docker_file(docker_file, temp_dir)

        logger.info(f"Building image at {temp_dir}")
        image = build_image(model_image_name, temp_dir)

        return image


def _copy_model(model_path, save_path):
    shutil.copytree(model_path, f"{save_path}/{model_path.name}")


def _copy_roheboam(save_path):
    roheboam_project_path = Path(
        importlib.util.find_spec("roheboam").origin
    ).parent.parent
    shutil.copytree(
        roheboam_project_path / "roheboam", f"{save_path}/roheboam/roheboam"
    )


def _copy_dependency_files(save_path, conda_env_file_path):
    roheboam_project_path = Path(
        importlib.util.find_spec("roheboam").origin
    ).parent.parent
    shutil.copy2(roheboam_project_path / "setup.py", f"{save_path}/setup.py")
    shutil.copy2(conda_env_file_path, f"{save_path}/environment.yml")


def _create_docker_file(model_path, model_format, build_env_image, dev_build):
    docker_template = dedent(
        f"""
        FROM {build_env_image}
        RUN apt-get update && DEBIAN_FRONTEND="noninteractive" apt-get install \
            'ffmpeg'\
            'libsm6' \
            'libxext6' \
            'libvips-dev' -y
        ENV SSL_CERT_DIR=/etc/ssl/certs
        """
    )

    if dev_build:
        docker_template += dedent(
            f"""
            WORKDIR /home/build/conda
            RUN conda create -n roheboam
            RUN conda install mamba -n base -c conda-forge
            RUN conda config --set channel_priority false
            COPY environment.yml /home/build/conda/environment.yml
            RUN mamba env update -n roheboam --file environment.yml
            ENV PATH="/opt/conda/envs/roheboam/bin:$PATH"
            """
        )

        docker_template += dedent(
            f"""
            COPY ./roheboam /home/build/roheboam
            COPY setup.py /home/build/roheboam/setup.py
            WORKDIR /home/build/roheboam
            RUN python setup.py install
            """
        )

    docker_template += dedent(
        f"""
            RUN mkdir -p /home/model
            WORKDIR /home/model
            COPY {model_path.name} /home/model
            RUN mkdir -p /home/.cache
            ENV TORCH_HOME=/home/.cache
            RUN chmod -R 777 /home
        """
    )

    if model_format == "MLFLOW":
        docker_template += dedent(
            f"""
           CMD exec roheboam serve --model_path /home/model --model_format mlflow
        """
        )

    if model_format == "SELDON":
        docker_template += dedent(
            f"""
            # Port for GRPC
            EXPOSE 5000
            # Port for REST
            EXPOSE 9000

            # Define environment variables
            ENV MODEL_NAME roheboam.engine.vision.SeldonImageModel
            ENV SERVICE_TYPE MODEL
            ENV SELDON_MODEL_PATH /home/model
            ENV FLASK_SINGLE_THREADED 1
            CMD exec roheboam serve --model_path /home/model --model_format seldon --debug
        """
        )
    return docker_template


def _write_docker_file(docker_file, save_path):
    with (Path(save_path) / "Dockerfile").open("w") as f:
        f.write(docker_file)



try:
    # TODO: remove if not supporting activeloop
    from .activeloop import *
    from .activeloop import lookup as activeloop_lookup
    from .roheboam import *
    from .roheboam import lookup as roheboam_lookup
    from .ultralytics import *
    from .ultralytics import lookup as ultralytics_lookup

    lookup = {**ultralytics_lookup, **activeloop_lookup, **roheboam_lookup}
except Exception:
    from .roheboam import *
    from .roheboam import lookup as roheboam_lookup
    from .ultralytics import *
    from .ultralytics import lookup as ultralytics_lookup

    lookup = {**ultralytics_lookup, **roheboam_lookup}



from abc import abstractmethod

import numpy as np

from ...utils.convenience import camel2snake
from ...utils.interfaces import MapDictionaryToMethodArguments


class LossWrapper:
    def __init__(self, losses):
        super().__init__()
        for loss in losses:
            setattr(self, camel2snake(loss.__class__.__name__), loss)
        self.losses = losses

    def __call__(self, predictions, targets={}, auxillary={}):
        total_loss = 0
        for loss in self.losses:
            loss_args = loss.to_call_kwargs(
                dict(predictions=predictions, targets=targets, auxillary=auxillary)
            )
            total_loss += loss(**loss_args).float()
        return total_loss

    @property
    def reduced_losses(self):
        losses = {}
        for loss in self.losses:
            try:
                per_batch_losses = loss.reduced_loss
                if isinstance(per_batch_losses, dict):
                    losses.update(loss.reduced_loss)
                else:
                    losses[loss.name] = loss.reduced_loss
            except AttributeError:
                losses[loss.name] = np.array([None])
        return losses

    @property
    def per_sample_losses(self):
        losses = {}
        for loss in self.losses:
            try:
                per_sample_loss = loss.per_sample_loss
                if isinstance(per_sample_loss, dict):
                    losses.update(loss.per_sample_loss)
                else:
                    losses[loss.name] = loss.per_sample_loss
            except AttributeError:
                losses[loss.name] = np.array([None])
        return losses

    @property
    def total_loss(self):
        return sum([loss.total_loss for loss in self.losses])


class BaseLoss(MapDictionaryToMethodArguments):
    loss_name_incrementor = 0

    def __init__(self, name):
        if name is None:
            self.name = f"{self.__class__.__name__}-{BaseLoss.loss_name_incrementor}"
            BaseLoss.loss_name_incrementor += 1
        else:
            self.name = name

    def to_call_kwargs(self, d):
        return d

    @staticmethod
    def reshape_to_batch_size_x_minus_one_and_sum_over_last_dimension(tensor):
        return BaseLoss.reshape_to_batch_size_x_minus_one_aggregate_over_last_dimension(
            tensor, aggregate_method="SUM"
        )

    @staticmethod
    def reshape_to_batch_size_x_minus_one_aggregate_over_last_dimension(
        tensor, aggregate_method="SUM"
    ):
        batch_size = tensor.size(0)
        if aggregate_method == "SUM":
            return tensor.view(batch_size, -1).sum(dim=1)
        elif aggregate_method == "MEAN":
            return tensor.view(batch_size, -1).mean(dim=1)
        else:
            raise KeyError("Aggregation method not found")

    @property
    @abstractmethod
    def unreduced_loss(self):
        pass

    @property
    @abstractmethod
    def per_sample_loss(self):
        pass

    @property
    @abstractmethod
    def reduced_loss(self):
        pass



from pathlib import Path
from uuid import uuid4

import matplotlib.pyplot as plt
import numpy as np

from ......core.data import Data
from ......vision.utils import imread_rgb


class ImageClassificationSample:
    def __init__(self, image, label=None, name=None):
        self.image = image
        self.label = label
        self.name = str(uuid4()) if name is None else name

    @classmethod
    def create(
        cls,
        image_data=None,
        image_path=None,
        load_image_fn=imread_rgb,
        label_data=None,
        label_path=None,
        load_label_fn=np.load,
        name=None,
    ):
        if name is None:
            name = Path(image_path).stem if image_path is not None else str(uuid4())

        image = Data(image_data, image_path, load_image_fn)
        if label_data is None and label_path is None:
            label = None
        else:
            label = Data(label_data, label_path, load_label_fn, should_cache=True)
        return cls(image, label, name)

    @property
    def has_label(self):
        return self.label is not None

    @property
    def data(self):
        if self.has_label:
            return self.image.data, self.label.data
        else:
            return self.image.data, None

    def plot(self, figsize=(8, 6)):
        plt.figure(figsize=figsize)
        plt.imshow(self.image.data)
        if self.has_label:
            plt.title(f"Class is: {self.label.data}")

    def create_plot_array(self):
        return self.image.data


lookup = {"ImageClassificationSample": ImageClassificationSample}



from . import BaseLoss


class TorchVisionFasterRCNNLoss(BaseLoss):
    def __init__(self, loss_key, name=None):
        super().__init__(name)
        self.loss_key = loss_key

    def to_call_kwargs(self, d):
        return {"loss": d["predictions"][self.loss_key]}

    @property
    def unreduced_loss(self):
        return self.loss

    @property
    def per_sample_loss(self):
        return self.loss

    @property
    def reduced_loss(self):
        return self.loss

    @property
    def total_loss(self):
        return self.loss

    def __call__(self, loss):
        self.loss = loss
        return loss
